# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.table import Table
from astropy.time import Time
import matplotlib.pyplot as plt
from gammapy.catalog.fermi import SourceCatalog3FGL, SourceCatalog4FGL
from gammapy.data import GTI
from gammapy.estimators import FluxPoints
from gammapy.estimators.map.core import DEFAULT_UNIT
from gammapy.estimators.utils import get_rebinned_axis
from gammapy.maps import MapAxis, RegionNDMap
from gammapy.modeling.models import PowerLawSpectralModel, SpectralModel
from gammapy.utils.scripts import make_path
from gammapy.utils.testing import (
    assert_quantity_allclose,
    mpl_plot_check,
    requires_data,
)
from gammapy.modeling.scipy import stat_profile_ul_scipy
from ..core import squash_fluxpoints

FLUX_POINTS_FILES = [
    "diff_flux_points.ecsv",
    "diff_flux_points.fits",
    "flux_points.ecsv",
    "flux_points.fits",
]


class LWTestModel(SpectralModel):
    @staticmethod
    def evaluate(x):
        return 1e4 * np.exp(-6 * x)

    def integral(self, xmin, xmax, **kwargs):
        return -1.0 / 6 * 1e4 * (np.exp(-6 * xmax) - np.exp(-6 * xmin))

    def inverse(self, y):
        return -1.0 / 6 * np.log(y * 1e-4)


class XSqrTestModel(SpectralModel):
    @staticmethod
    def evaluate(x):
        return x**2

    def integral(self, xmin, xmax, **kwargs):
        return 1.0 / 3 * (xmax**3 - xmin**2)

    def inverse(self, y):
        return np.sqrt(y)


class ExpTestModel(SpectralModel):
    @staticmethod
    def evaluate(x):
        return np.exp(x * u.Unit("1 / TeV"))

    def integral(self, xmin, xmax, **kwargs):
        return np.exp(xmax * u.Unit("1 / TeV")) - np.exp(xmin * u.Unit("1 / TeV"))

    def inverse(self, y):
        return np.log(y * u.TeV) * u.TeV


def test_energy_ref_lafferty():
    """
    Tests Lafferty & Wyatt x-point method.

    Using input function g(x) = 10^4 exp(-6x) against
    check values from paper Lafferty & Wyatt. Nucl. Instr. and Meth. in Phys.
    Res. A 355 (1995) 541-547, p. 542 Table 1
    """
    # These are the results from the paper
    desired = np.array([0.048, 0.190, 0.428, 0.762])

    model = LWTestModel()
    energy_min = np.array([0.0, 0.1, 0.3, 0.6])
    energy_max = np.array([0.1, 0.3, 0.6, 1.0])
    actual = FluxPoints._energy_ref_lafferty(model, energy_min, energy_max)
    assert_allclose(actual, desired, atol=1e-3)


@pytest.mark.xfail
def test_dnde_from_flux():
    """Tests y-value normalization adjustment method."""
    table = Table()
    table["e_min"] = np.array([10, 20, 30, 40])
    table["e_max"] = np.array([20, 30, 40, 50])
    table["flux"] = np.array([42, 52, 62, 72])  # 'True' integral flux in this test bin

    # Get values
    model = XSqrTestModel()
    table["e_ref"] = FluxPoints._energy_ref_lafferty(
        model, table["e_min"], table["e_max"]
    )
    dnde = FluxPoints.from_table(table, reference_model=model)

    # Set up test case comparison
    dnde_model = model(table["e_ref"])

    # Test comparison result
    desired = model.integral(table["e_min"], table["e_max"])
    # Test output result
    actual = table["flux"] * (dnde_model / dnde)
    # Compare
    assert_allclose(actual, desired, rtol=1e-6)


@pytest.mark.xfail
@pytest.mark.parametrize("method", ["table", "lafferty", "log_center"])
def test_compute_flux_points_dnde_exp(method):
    """
    Tests against analytical result or result from a powerlaw.
    """
    model = ExpTestModel()

    energy_min = [1.0, 10.0] * u.TeV
    energy_max = [10.0, 100.0] * u.TeV

    table = Table()
    table.meta["SED_TYPE"] = "flux"
    table["e_min"] = energy_min
    table["e_max"] = energy_max

    flux = model.integral(energy_min, energy_max)
    table["flux"] = flux

    if method == "log_center":
        energy_ref = np.sqrt(energy_min * energy_max)
    elif method == "table":
        energy_ref = [2.0, 20.0] * u.TeV
    elif method == "lafferty":
        energy_ref = FluxPoints._energy_ref_lafferty(model, energy_min, energy_max)

    table["e_ref"] = energy_ref

    result = FluxPoints.from_table(table, reference_model=model)

    # Test energy
    actual = result.energy_ref
    assert_quantity_allclose(actual, energy_ref, rtol=1e-8)

    # Test flux
    actual = result.dnde
    desired = model(energy_ref)
    assert_quantity_allclose(actual, desired, rtol=1e-8)


@requires_data()
def test_fermi_to_dnde():
    catalog_4fgl = SourceCatalog4FGL("$GAMMAPY_DATA/catalogs/fermi/gll_psc_v20.fit.gz")
    src = catalog_4fgl["FGES J1553.8-5325"]
    fp = src.flux_points

    assert_allclose(
        fp.dnde.quantity[1, 0, 0],
        4.567393e-10 * u.Unit("cm-2 s-1 MeV-1"),
        rtol=1e-5,
    )


@pytest.fixture(params=FLUX_POINTS_FILES, scope="session")
def flux_points(request):
    path = "$GAMMAPY_DATA/tests/spectrum/flux_points/" + request.param
    return FluxPoints.read(path)


@pytest.fixture(scope="session")
def flux_points_likelihood():
    path = "$GAMMAPY_DATA/estimators/crab_hess_fp/crab_hess_fp.fits"
    return FluxPoints.read(path)


@requires_data()
class TestFluxPoints:
    def test_info(self, flux_points):
        info = str(flux_points)
        assert "geom" in info
        assert "axes" in info
        assert "ref. model" in info
        assert "quantities" in info

    def test_energy_ref(self, flux_points):
        actual = flux_points.energy_ref
        desired = np.sqrt(flux_points.energy_min * flux_points.energy_max)
        assert_quantity_allclose(actual, desired)

    def test_energy_min(self, flux_points):
        actual = flux_points.energy_min
        desired = 299530.97 * u.MeV
        assert_quantity_allclose(actual.sum(), desired)

    def test_energy_max(self, flux_points):
        actual = flux_points.energy_max
        desired = 399430.975 * u.MeV
        assert_quantity_allclose(actual.sum(), desired)

    def test_write_fits(self, tmp_path, flux_points):
        start = u.Quantity([1, 2], "min")
        stop = u.Quantity([1.5, 2.5], "min")
        time_ref = Time("2010-01-01 00:00:00.0")
        gti = GTI.create(start, stop, time_ref)
        flux_points.gti = gti
        flux_points.write(tmp_path / "tmp.fits", sed_type=flux_points.sed_type_init)
        actual = FluxPoints.read(tmp_path / "tmp.fits")
        actual._data.pop("is_ul", None)
        flux_points._data.pop("is_ul", None)
        assert actual.gti.time_start[0] == gti.time_start[0]
        assert str(flux_points) == str(actual)

    def test_write_ecsv(self, tmp_path, flux_points):
        filename = tmp_path / "flux_points.ecsv"
        filename.touch()
        flux_points.write(
            filename,
            sed_type=flux_points.sed_type_init,
            overwrite=True,
        )
        actual = FluxPoints.read(filename)
        actual._data.pop("is_ul", None)
        flux_points._data.pop("is_ul", None)
        assert str(flux_points) == str(actual)

    def test_quantity_access(self, flux_points_likelihood):
        assert flux_points_likelihood.sqrt_ts
        assert flux_points_likelihood.ts
        assert flux_points_likelihood.stat
        assert_allclose(flux_points_likelihood.n_sigma_ul, 3)
        assert flux_points_likelihood.sed_type_init == "likelihood"

    def test_plot(self, flux_points):
        fig = plt.figure()
        ax = fig.add_axes([0.2, 0.2, 0.7, 0.7])
        ax.xaxis.set_units(u.eV)

        yunit = DEFAULT_UNIT[flux_points.sed_type_init]
        ax.yaxis.set_units(yunit)

        with mpl_plot_check():
            flux_points.plot(ax=ax)

    def test_plot_likelihood(self, flux_points_likelihood):
        plt.figure()
        with mpl_plot_check():
            flux_points_likelihood.plot_ts_profiles()

    def test_plot_likelihood_error(self, flux_points_likelihood):
        del flux_points_likelihood._data["stat_scan"]

        with pytest.raises(AttributeError):
            fig = plt.figure()
            ax = fig.add_subplot()
            flux_points_likelihood.plot_ts_profiles(ax=ax)


@requires_data()
def test_plot_format_yaxis():
    path = "$GAMMAPY_DATA/tests/spectrum/flux_points/flux_points.fits"
    fp = FluxPoints.read(path)

    with mpl_plot_check():
        ax = fp.plot(sed_type="e2dnde")
        assert (
            ax.yaxis.get_label().get_text()
            == "e2dnde [$\\mathrm{erg\\,s^{-1}\\,cm^{-2}}$]"
        )

    with mpl_plot_check():
        ax = fp.plot(sed_type="dnde", energy_power=2)
        assert (
            ax.yaxis.get_label().get_text()
            == "e2 * dnde [$\\mathrm{TeV\\,s^{-1}\\,cm^{-2}}$]"
        )

    with mpl_plot_check():
        ax = fp.plot(sed_type="dnde")
        assert (
            ax.yaxis.get_label().get_text()
            == "dnde [$\\mathrm{TeV^{-1}\\,s^{-1}\\,cm^{-2}}$]"
        )

    with mpl_plot_check():
        ax = fp.plot(sed_type="dnde", energy_power=2.7)
        assert (
            ax.yaxis.get_label().get_text()
            == "e2.7 * dnde [$\\mathrm{TeV^{17/10}\\,s^{-1}\\,cm^{-2}}$]"
        )


@requires_data()
def test_flux_points_single_bin_dnde():
    path = make_path("$GAMMAPY_DATA/tests/spectrum/flux_points/diff_flux_points.fits")
    table = Table.read(path)

    table_single_bin = table[1:2]
    fp = FluxPoints.from_table(table_single_bin, sed_type="dnde")

    with pytest.raises(ValueError):
        _ = fp.flux_ref

    with mpl_plot_check():
        fp.plot(sed_type="e2dnde")

    with pytest.raises(ValueError):
        fp.to_table(sed_type="flux")

    table = fp.to_table(sed_type="dnde")

    assert_allclose(table["e_ref"], 153.992 * u.MeV, rtol=0.001)
    assert "e_min" not in table.colnames
    assert "e_max" not in table.colnames


@requires_data()
def test_compute_flux_points_dnde_fermi():
    """
    Test compute_flux_points_dnde on fermi source.
    """
    fermi_3fgl = SourceCatalog3FGL()
    source = fermi_3fgl["3FGL J0835.3-4510"]
    flux_points = source.flux_points
    table = source.flux_points_table

    for column in ["e2dnde", "e2dnde_errn", "e2dnde_errp", "e2dnde_ul"]:
        actual = table[column].quantity
        desired = getattr(flux_points, column).quantity.squeeze()
        assert_quantity_allclose(actual[:-1], desired[:-1], rtol=0.05)


@requires_data()
def test_plot_fp_no_ul():
    path = make_path("$GAMMAPY_DATA/tests/spectrum/flux_points/diff_flux_points.fits")
    table = Table.read(path)
    table.remove_column("dnde_ul")
    fp = FluxPoints.from_table(table, sed_type="dnde")

    with mpl_plot_check():
        fp.plot()


@requires_data()
def test_is_ul(tmp_path):
    catalog_4fgl = SourceCatalog4FGL("$GAMMAPY_DATA/catalogs/fermi/gll_psc_v20.fit.gz")
    src = catalog_4fgl["FGES J1553.8-5325"]
    fp = src.flux_points

    is_ul = fp._data["is_ul"].data.squeeze()

    assert_allclose(fp.is_ul.data.squeeze(), is_ul)
    table = fp.to_table()
    assert_allclose(table["is_ul"].data.data, is_ul)

    fp.sqrt_ts_threshold_ul = 100
    assert_allclose(fp.is_ul.data.squeeze(), np.ones(is_ul.shape, dtype=bool))
    table = fp.to_table()
    assert_allclose(table["is_ul"].data.data, np.ones(is_ul.shape, dtype=bool))

    table.write(tmp_path / "test_modif_ul_threshold.fits")
    table_read = Table.read(tmp_path / "test_modif_ul_threshold.fits")
    assert_allclose(table_read["is_ul"].data.data, np.ones(is_ul.shape, dtype=bool))
    fp_read = FluxPoints.from_table(table_read)
    assert_allclose(fp_read.is_ul.data.squeeze(), np.ones(is_ul.shape, dtype=bool))
    assert_allclose(fp_read.to_table()["is_ul"], fp_read.is_ul.data.squeeze())

    fp.is_ul = is_ul
    assert_allclose(fp.is_ul.data.squeeze(), is_ul)
    table = fp.to_table()
    assert_allclose(table["is_ul"].data.data, is_ul)


def test_flux_points_plot_no_error_bar():
    table = Table()
    pwl = PowerLawSpectralModel()
    e_ref = np.geomspace(1, 100, 7) * u.TeV

    table["e_ref"] = e_ref
    table["dnde"] = pwl(e_ref)
    table.meta["SED_TYPE"] = "dnde"

    flux_points = FluxPoints.from_table(table)
    with mpl_plot_check():
        _ = flux_points.plot(sed_type="dnde")


@requires_data()
def test_fp_no_is_ul():
    path = make_path("$GAMMAPY_DATA/tests/spectrum/flux_points/flux_points.fits")
    table = Table.read(path)
    table.remove_column("is_ul")
    table.remove_column("flux_ul")

    fp = FluxPoints.from_table(table)
    fp_table = fp.to_table()
    assert "is_ul" not in fp_table.colnames


def test_table_columns():
    table = Table()
    table["e_min"] = np.array([10, 20, 30, 40]) * u.TeV
    table["e_max"] = np.array([20, 30, 40, 50]) * u.TeV
    table["flux"] = np.array([42, 52, 62, 72]) / (u.s * u.cm * u.cm)
    table["other"] = np.array([1, 2, 3, 4])
    table["n_dof"] = np.array([1, 2, 1, 2])

    # Get values
    model = PowerLawSpectralModel()
    table["e_ref"] = FluxPoints._energy_ref_lafferty(
        model, table["e_min"], table["e_max"]
    )
    fp = FluxPoints.from_table(table, reference_model=model)

    assert fp.available_quantities == ["norm", "n_dof"]
    assert_allclose(fp.n_dof.data.ravel(), table["n_dof"])


@requires_data()
def test_resample_axis():
    lc_1d = FluxPoints.read(
        "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
        format="lightcurve",
    )
    axis_new = get_rebinned_axis(
        lc_1d, method="fixed-bins", group_size=5, axis_name="time"
    )
    l1 = lc_1d.resample_axis(axis_new=axis_new)
    assert_allclose(l1.norm.data.ravel()[0:2], [1.56321943, 2.12845751], rtol=1e-3)
    assert_allclose(l1.norm_err.data.ravel()[0:2], [0.03904136, 0.03977413], rtol=1e-3)
    assert_allclose(l1.n_dof.data.ravel()[0], 5.0)
    assert l1.success.data.ravel()[0]

    axis_new = get_rebinned_axis(
        lc_1d, method="min-ts", ts_threshold=300, axis_name="time"
    )
    l1 = lc_1d.resample_axis(axis_new=axis_new)
    assert_allclose(l1.norm_err.data.ravel()[0:2], [0.072236, 0.092942], rtol=1e-3)
    assert_allclose(l1.ts.data.ravel()[0:2], [312.742222, 454.99609], rtol=1e-3)
    assert_allclose(l1.stat_null.data.ravel()[0:2], [319.8675, 462.329], rtol=1e-3)
    assert l1.success.data.ravel()[0]
    assert_allclose(l1.n_dof.data[0][0][0][0], 2)

    path = make_path("$GAMMAPY_DATA/tests/spectrum/flux_points/flux_points.fits")
    table = Table.read(path)
    fp = FluxPoints.from_table(table)
    with pytest.raises(ValueError):
        fp.resample_axis(axis_new=MapAxis.from_nodes([0, 1, 2]))


def test_flux_point_init():
    table = Table()
    table["norm"] = np.array([10, 20, 30, 40])
    table["e_min"] = np.array([10, 20, 30, 40]) * u.TeV
    table["e_max"] = np.array([20, 30, 40, 50]) * u.TeV
    norm_map = RegionNDMap.from_table(table=table, colname="norm", format="gadf-sed")

    maps = dict(norm=norm_map)
    # Get values
    model = PowerLawSpectralModel()
    table["e_ref"] = FluxPoints._energy_ref_lafferty(
        model, table["e_min"], table["e_max"]
    )
    fp = FluxPoints(maps, reference_model=model)

    assert fp.available_quantities == ["norm"]
    assert_allclose(fp.norm.data.ravel(), [10, 20, 30, 40])


@requires_data()
def test_recompute_ul():
    flux_points = FluxPoints.read(
        "$GAMMAPY_DATA/estimators/crab_hess_fp/crab_hess_fp.fits"
    )
    value_scan = flux_points.stat_scan.geom.axes["norm"].center
    idx = 0
    stat_scan = flux_points.stat_scan.data[idx].squeeze()

    with pytest.raises(
        ValueError,
        match="Statistic profile has no finite value therefore no best-fit value can be determined.",
    ):
        stat_profile_ul_scipy(value_scan, stat_scan, delta_ts=4)

    # As we have the above error, the recompute_ul should set this first value to NaN
    recomputed = flux_points.recompute_ul()
    assert_allclose(recomputed.norm_ul.data[0][0], np.nan)


@requires_data()
def test_squash_fluxpoints():
    fp1 = FluxPoints.read("$GAMMAPY_DATA/estimators/crab_hess_fp/crab_hess_fp.fits")
    energy_axis = fp1.geom.axes["energy"]
    squash_fp1 = squash_fluxpoints(flux_point=fp1, axis=energy_axis)
    assert squash_fp1.geom.axes.names == []
    assert len(fp1.stat.data) == 5
    assert len(squash_fp1.stat.data) == 1

    fp2 = FluxPoints.read(
        "$GAMMAPY_DATA/estimators/pks2155_hess_lc/pks2155_hess_lc.fits",
        format="lightcurve",
    )
    time_axis = fp2.geom.axes["time"]
    squash_fp2 = squash_fluxpoints(fp2, axis=time_axis)
    assert squash_fp2.geom.axes["energy"].nbin == 1
    assert len(fp2.stat.data) == 34
    assert len(squash_fp2.stat.data) == 1



"""
Models
======
This is an introduction and overview on how to work with models in
Gammapy.

The sub-package `~gammapy.modeling` contains all the functionality
related to modeling and fitting data. This includes spectral, spatial
and temporal model classes, as well as the fit and parameter API.The
models follow a naming scheme which contains the category as a suffix to
the class name. An overview of all the available models can be found in
the :ref:`model-gallery`.

Note that there are separate tutorials,
:doc:`/tutorials/details/model_management` and
:doc:`/tutorials/details/fitting` that explains about
`~gammapy.modeling`, the Gammapy modeling and fitting framework. You
have to read that to learn how to work with models in order to analyse
data.

"""

######################################################################
# Setup
# -----
#

# %matplotlib inline
import numpy as np
from astropy import units as u
import matplotlib.pyplot as plt
from IPython.display import display
from gammapy.maps import Map, MapAxis, WcsGeom


######################################################################
# Spectral models
# ---------------
#
# All models are imported from the `~gammapy.modeling.models` namespace.
# Let’s start with a `~gammapy.modeling.models.PowerLawSpectralModel`:
#

from gammapy.modeling.models import PowerLawSpectralModel

pwl = PowerLawSpectralModel()
print(pwl)


######################################################################
# To get a list of all available spectral models you can import and print
# the spectral model registry or take a look at the :ref:`model-gallery`
#

from gammapy.modeling.models import SPECTRAL_MODEL_REGISTRY

print(SPECTRAL_MODEL_REGISTRY)


######################################################################
# Spectral models all come with default parameters. Different parameter
# values can be passed on creation of the model, either as a string
# defining the value and unit or as an `astropy.units.Quantity` object
# directly:
#

amplitude = 1e-12 * u.Unit("TeV-1 cm-2 s-1")
pwl = PowerLawSpectralModel(amplitude=amplitude, index=2.2)


######################################################################
# For convenience a `str` specifying the value and unit can be passed as
# well:
#

pwl = PowerLawSpectralModel(amplitude="2.7e-12 TeV-1 cm-2 s-1", index=2.2)
print(pwl)


######################################################################
# The model can be evaluated at given energies by calling the model
# instance:
#

energy = [1, 3, 10, 30] * u.TeV
dnde = pwl(energy)
print(dnde)


######################################################################
# The returned quantity is a differential photon flux.
#
# For spectral models you can additionally compute the integrated and
# energy flux in a given energy range:
#

flux = pwl.integral(energy_min=1 * u.TeV, energy_max=10 * u.TeV)
print(flux)

eflux = pwl.energy_flux(energy_min=1 * u.TeV, energy_max=10 * u.TeV)
print(eflux)


######################################################################
# This also works for a list or an array of integration boundaries:
#

energy = [1, 3, 10, 30] * u.TeV
flux = pwl.integral(energy_min=energy[:-1], energy_max=energy[1:])
print(flux)


######################################################################
# In some cases it can be useful to find use the inverse of a spectral
# model, to find the energy at which a given flux is reached:
#

dnde = 2.7e-12 * u.Unit("TeV-1 cm-2 s-1")
energy = pwl.inverse(dnde)
print(energy)


######################################################################
# As a convenience you can also plot any spectral model in a given energy
# range:
#

pwl.plot(energy_bounds=[1, 100] * u.TeV)
plt.show()


######################################################################
# Norm Spectral Models
# ~~~~~~~~~~~~~~~~~~~~
#


######################################################################
# Normed spectral models are a special class of Spectral Models, which
# have a dimension-less normalisation. These spectral models feature a
# norm parameter instead of amplitude and are named using the
# ``NormSpectralModel`` suffix. They **must** be used along with another
# spectral model, as a multiplicative correction factor according to their
# spectral shape. They can be typically used for adjusting template based
# models, or adding a EBL correction to some analytic model.
#
# To check if a given `~gammapy.modeling.models.SpectralModel` is a norm model, you can simply
# look at the ``is_norm_spectral_model`` property
#

# To see the available norm models shipped with gammapy:
for model in SPECTRAL_MODEL_REGISTRY:
    if model.is_norm_spectral_model:
        print(model)


######################################################################
# As an example, we see the `~gammapy.modeling.models.PowerLawNormSpectralModel`
#

from gammapy.modeling.models import PowerLawNormSpectralModel

pwl_norm = PowerLawNormSpectralModel(tilt=0.1)
print(pwl_norm)


######################################################################
# We can check the correction introduced at each energy
#

energy = [0.3, 1, 3, 10, 30] * u.TeV
print(pwl_norm(energy))


######################################################################
# A typical use case of a norm model would be in applying spectral
# correction to a `~gammapy.modeling.models.TemplateSpectralModel`. A template model is defined
# by custom tabular values provided at initialization.
#

from gammapy.modeling.models import TemplateSpectralModel

energy = [0.3, 1, 3, 10, 30] * u.TeV
values = [40, 30, 20, 10, 1] * u.Unit("TeV-1 s-1 cm-2")
template = TemplateSpectralModel(energy, values)
template.plot(energy_bounds=[0.2, 50] * u.TeV, label="template model")
normed_template = template * pwl_norm
normed_template.plot(energy_bounds=[0.2, 50] * u.TeV, label="normed_template model")
plt.legend()
plt.show()


######################################################################
# Compound Spectral Model
# ~~~~~~~~~~~~~~~~~~~~~~~
#
# A `~gammapy.modeling.models.CompoundSpectralModel` is an arithmetic combination of two spectral
# models. The model ``normed_template`` created in the preceding example
# is an example of a `~gammapy.modeling.models.CompoundSpectralModel`
#

print(normed_template)


######################################################################
# To create an additive model, you can do simply:
#

model_add = pwl + template
print(model_add)


######################################################################
# Spatial models
# --------------
#


######################################################################
# Spatial models are imported from the same `~gammapy.modeling.models`
# namespace, let’s start with a `~gammapy.modeling.models.GaussianSpatialModel`:
#

from gammapy.modeling.models import GaussianSpatialModel

gauss = GaussianSpatialModel(lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg")
print(gauss)


######################################################################
# Again you can check the ``SPATIAL_MODELS`` registry to see which models
# are available or take a look at the :ref:`model-gallery`
#

from gammapy.modeling.models import SPATIAL_MODEL_REGISTRY

print(SPATIAL_MODEL_REGISTRY)


######################################################################
# The default coordinate frame for all spatial models is ``"icrs"``, but
# the frame can be modified using the ``frame`` argument:
#

gauss = GaussianSpatialModel(
    lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg", frame="galactic"
)


######################################################################
# You can specify any valid `astropy.coordinates` frame. The center
# position of the model can be retrieved as a
# `astropy.coordinates.SkyCoord` object using `~gammapy.modeling.models.SpatialModel.position`:
#

print(gauss.position)


######################################################################
# Spatial models can be evaluated again by calling the instance:
#

lon = [0, 0.1] * u.deg
lat = [0, 0.1] * u.deg

flux_per_omega = gauss(lon, lat)
print(flux_per_omega)


######################################################################
# The returned quantity corresponds to a surface brightness. Spatial model
# can be also evaluated using `~gammapy.maps.Map` and
# `~gammapy.maps.Geom` objects:
#

m = Map.create(skydir=(0, 0), width=(1, 1), binsz=0.02, frame="galactic")
m.quantity = gauss.evaluate_geom(m.geom)
m.plot(add_cbar=True)
plt.show()


######################################################################
# Again for convenience the model can be plotted directly:
#
gauss.plot(add_cbar=True)
plt.show()


######################################################################
# All spatial models have an associated sky region to it e.g. to
# illustrate the extension of the model on a sky image. The returned object
# is an `~regions.SkyRegion` object:
#

print(gauss.to_region())


######################################################################
# Now we can plot the region on a sky image:
#

plt.figure()
gauss_elongated = GaussianSpatialModel(
    lon_0="0 deg", lat_0="0 deg", sigma="0.2 deg", e=0.7, phi="45 deg"
)
ax = gauss_elongated.plot(add_cbar=True)

region = gauss_elongated.to_region()
region_pix = region.to_pixel(ax.wcs)
ax.add_artist(region_pix.as_artist(ec="w", fc="None"))
plt.show()


######################################################################
# The `~gammapy.modeling.models.GaussianSpatialModel.to_region()` method can also be useful to write e.g. ds9 region
# files using ``write_ds9`` from the `regions` package:
#

from regions import Regions

regions = Regions([gauss.to_region(), gauss_elongated.to_region()])

filename = "regions.reg"
regions.write(
    filename,
    format="ds9",
    overwrite=True,
)

# !cat regions.reg


######################################################################
# Temporal models
# ---------------
#


######################################################################
# Temporal models are imported from the same `~gammapy.modeling.models`
# namespace, let’s start with a `~gammapy.modeling.models.GaussianTemporalModel`:
#

from gammapy.modeling.models import GaussianTemporalModel

gauss_temp = GaussianTemporalModel(t_ref=59240.0 * u.d, sigma=2.0 * u.d)
print(gauss_temp)


######################################################################
# To check the ``TEMPORAL_MODELS`` registry to see which models are
# available:
#

from gammapy.modeling.models import TEMPORAL_MODEL_REGISTRY

print(TEMPORAL_MODEL_REGISTRY)


######################################################################
# Temporal models can be evaluated on `astropy.time.Time` objects. The
# returned quantity is a dimensionless number
#

from astropy.time import Time

time = Time("2021-01-29 00:00:00.000")
gauss_temp(time)


######################################################################
# As for other models, they can be plotted in a given time range
#

time = Time([59233.0, 59250], format="mjd")
gauss_temp.plot(time)
plt.show()


######################################################################
# `~gammapy.modeling.models.SkyModel`
# -----------------------------------
#
# The `~gammapy.modeling.models.SkyModel` class combines a spectral, and
# optionally, a spatial model and a temporal. It can be created from
# existing spectral, spatial and temporal model components:
#

from gammapy.modeling.models import SkyModel

model = SkyModel(
    spectral_model=pwl,
    spatial_model=gauss,
    temporal_model=gauss_temp,
    name="my-source",
)
print(model)


######################################################################
# It is good practice to specify a name for your sky model, so that you
# can access it later by name and have meaningful identifier you
# serialisation. If you don’t define a name, a unique random name is
# generated:
#

model_without_name = SkyModel(spectral_model=pwl, spatial_model=gauss)
print(model_without_name.name)


######################################################################
# The individual components of the source model can be accessed using
# ``.spectral_model``, ``.spatial_model`` and ``.temporal_model``:
#

print(model.spectral_model)

# %%
print(model.spatial_model)

# %%
print(model.temporal_model)


######################################################################
# And can be used as you have seen already seen above:
#

model.spectral_model.plot(energy_bounds=[1, 10] * u.TeV)
plt.show()


######################################################################
# Note that the gammapy fitting can interface only with a `~gammapy.modeling.models.SkyModel` and
# **not** its individual components. So, it is customary to work with
# `~gammapy.modeling.models.SkyModel` even if you are not doing a 3D fit. Since the amplitude
# parameter resides on the `~gammapy.modeling.models.SpectralModel`, specifying a spectral
# component is compulsory. The temporal and spatial components are
# optional. The temporal model needs to be specified only for timing
# analysis. In some cases (e.g. when doing a spectral analysis) there is
# no need for a spatial component either, and only a spectral model is
# associated with the source.
#

model_spectrum = SkyModel(spectral_model=pwl, name="source-spectrum")
print(model_spectrum)


######################################################################
# Additionally the spatial model of `~gammapy.modeling.models.SkyModel`
# can be used to represent source models based on templates, where the
# spatial and energy axes are correlated. It can be created e.g. from an
# existing FITS file:
#

from gammapy.modeling.models import PowerLawNormSpectralModel, TemplateSpatialModel

diffuse_cube = TemplateSpatialModel.read(
    "$GAMMAPY_DATA/fermi-3fhl-gc/gll_iem_v06_gc.fits.gz", normalize=False
)
diffuse = SkyModel(PowerLawNormSpectralModel(), diffuse_cube)
print(diffuse)


######################################################################
# Note that if the spatial model is not normalized over the sky it has to
# be combined with a normalized spectral model, for example
# `~gammapy.modeling.models.PowerLawNormSpectralModel`. This is the only
# case in `~gammapy.modeling.models.SkyModel` where the unit is fully attached to
# the spatial model.
#


######################################################################
# .. _modifying-model-parameters:
#
# Modifying model parameters
# --------------------------
#
# Model parameters can be modified (eg: frozen, values changed, etc at any
# point), eg:
#

# Freezing a parameter
model.spectral_model.index.frozen = True
# Making a parameter free
model.spectral_model.index.frozen = False

# Changing a value
model.spectral_model.index.value = 3

# Setting min and max ranges on parameters
model.spectral_model.index.min = 1.0
model.spectral_model.index.max = 5.0

# Visualise the model as a table
display(model.parameters.to_table())


######################################################################
# You can use the interactive boxes to choose model parameters by name,
# type or other attributes mentioned in the column names.
#


######################################################################
# Model lists and serialisation
# -----------------------------
#
# In a typical analysis scenario a model consists of multiple model
# components, or a “catalog” or “source library”. To handle this list of
# multiple model components, Gammapy has a `~gammapy.modeling.models.Models` class:
#

from gammapy.modeling.models import Models

models = Models([model, diffuse])
print(models)


######################################################################
# Individual model components in the list can be accessed by their name:
#

print(models["my-source"])


######################################################################
# **Note:** To make the access by name unambiguous, models are required to
# have a unique name, otherwise an error will be thrown.
#
# To see which models are available you can use the ``.names`` attribute:
#

print(models.names)


######################################################################
# Note that a `~gammapy.modeling.models.SkyModel` object can be evaluated for a given longitude,
# latitude, and energy, but the `~gammapy.modeling.models.Models` object cannot.
# This `~gammapy.modeling.models.Models`
# container object will be assigned to `~gammapy.datasets.Dataset` or `~gammapy.datasets.Datasets`
# together with the data to be fitted. Checkout e.g. the
# :doc:`/tutorials/details/model_management` tutorial for details.
#
# The `~gammapy.modeling.models.Models` class also has in place ``.append()`` and ``.extend()``
# methods:
#

model_copy = model.copy(name="my-source-copy")
models.append(model_copy)


######################################################################
# This list of models can be also serialised to a custom YAML based
# format:
#

models_yaml = models.to_yaml()
print(models_yaml)


######################################################################
# The structure of the yaml files follows the structure of the python
# objects. The ``components`` listed correspond to the `~gammapy.modeling.models.SkyModel` and
# components of the `~gammapy.modeling.models.Models`. For each `~gammapy.modeling.models.SkyModel`
# we have information about its ``name``, ``type`` (corresponding to the
# tag attribute) and sub-models (i.e ``spectral`` model and eventually
# ``spatial`` model). Then the spatial and spectral models are defined by
# their type and parameters. The ``parameters`` keys name/value/unit are
# mandatory, while the keys min/max/frozen are optional (so you can
# prepare shorter files).
#
# If you want to write this list of models to disk and read it back later
# you can use:
#

models.write("models.yaml", overwrite=True)

models_read = Models.read("models.yaml")


######################################################################
# Additionally the models can be exported and imported together with the data
# using the `~gammapy.datasets.Datasets.read()` and `~gammapy.datasets.Datasets.write()` methods as shown
# in the :doc:`/tutorials/analysis-3d/analysis_mwl`
# notebook.
#
# Models with shared parameter
# ----------------------------
#
# A model parameter can be shared with other models, for example we can
# define two power-law models with the same spectral index but different
# amplitudes:
#

pwl2 = PowerLawSpectralModel()
pwl2.index = pwl.index
pwl.index.value = (
    2.3  # also update pwl2 as the parameter object is now the same as shown below
)
print(pwl.index)
print(pwl2.index)


######################################################################
# In the YAML files the shared parameter is flagged by the additional
# ``link`` entry that follows the convention ``parameter.name@unique_id``:
#

models = Models([SkyModel(pwl, name="source1"), SkyModel(pwl2, name="source2")])
models_yaml = models.to_yaml()
print(models_yaml)


######################################################################
# .. _custom-model:
#
# Implementing a custom model
# ---------------------------
#
# In order to add a user defined spectral model you have to create a
# `~gammapy.modeling.models.SpectralModel` subclass. This new model class should include:
#
# -  a tag used for serialization (it can be the same as the class name)
# -  an instantiation of each Parameter with their unit, default values
#    and frozen status
# -  the evaluate function where the mathematical expression for the model
#    is defined.
#
# As an example we will use a PowerLawSpectralModel plus a Gaussian (with
# fixed width). First we define the new custom model class that we name
# ``MyCustomSpectralModel``:
#

from gammapy.modeling import Parameter
from gammapy.modeling.models import SpectralModel


class MyCustomSpectralModel(SpectralModel):
    """My custom spectral model, parametrizing a power law plus a Gaussian spectral line.

    Parameters
    ----------
    amplitude : `astropy.units.Quantity`
        Amplitude of the spectra model.
    index : `astropy.units.Quantity`
        Spectral index of the model.
    reference : `astropy.units.Quantity`
        Reference energy of the power law.
    mean : `astropy.units.Quantity`
        Mean value of the Gaussian.
    width : `astropy.units.Quantity`
        Sigma width of the Gaussian line.

    """

    tag = "MyCustomSpectralModel"
    amplitude = Parameter("amplitude", "1e-12 cm-2 s-1 TeV-1", min=0)
    index = Parameter("index", 2, min=0)
    reference = Parameter("reference", "1 TeV", frozen=True)
    mean = Parameter("mean", "1 TeV", min=0)
    width = Parameter("width", "0.1 TeV", min=0, frozen=True)

    @staticmethod
    def evaluate(energy, index, amplitude, reference, mean, width):
        pwl = PowerLawSpectralModel.evaluate(
            energy=energy,
            index=index,
            amplitude=amplitude,
            reference=reference,
        )
        gauss = amplitude * np.exp(-((energy - mean) ** 2) / (2 * width**2))
        return pwl + gauss


######################################################################
# It is good practice to also implement a docstring for the model,
# defining the parameters and also defining a ``.tag``, which specifies the
# name of the model for serialisation. Also note that gammapy assumes that
# all `~gammapy.modeling.models.SpectralModel` evaluate functions return a flux in unit of
# `"cm-2 s-1 TeV-1"` (or equivalent dimensions).
#
# This model can now be used as any other spectral model in Gammapy:
#

my_custom_model = MyCustomSpectralModel(mean="3 TeV")
print(my_custom_model)

print(my_custom_model.integral(1 * u.TeV, 10 * u.TeV))

my_custom_model.plot(energy_bounds=[1, 10] * u.TeV)
plt.show()


######################################################################
# As a next step we can also register the custom model in the
# ``SPECTRAL_MODELS`` registry, so that it becomes available for
# serialization:
#

SPECTRAL_MODEL_REGISTRY.append(MyCustomSpectralModel)

model = SkyModel(spectral_model=my_custom_model, name="my-source")
models = Models([model])
models.write("my-custom-models.yaml", overwrite=True)

# !cat my-custom-models.yaml


######################################################################
# Similarly you can also create custom spatial models and add them to the
# ``SPATIAL_MODELS`` registry. In that case gammapy assumes that the
# evaluate function return a normalized quantity in “sr-1” such as the
# model integral over the whole sky is one.
#


######################################################################
# Models with energy dependent morphology
# ---------------------------------------
#
# A common science case in the study of extended sources is to probe for
# energy dependent morphology, eg: in Supernova Remnants or Pulsar Wind
# Nebulae. Traditionally, this has been done by splitting the data into
# energy bands and doing individual fits of the morphology in these energy
# bands.
#
# `~gammapy.modeling.models.SkyModel` offers a natural framework to simultaneously model the
# energy and morphology, e.g. spatial extent described by a parametric
# model expression with energy dependent parameters.
#
# The models shipped within gammapy use a “factorised” representation of
# the source model, where the spatial (:math:`l,b`), energy (:math:`E`)
# and time (:math:`t`) dependence are independent model components and not
# correlated:
#
# .. math::
#     \begin{align}f(l, b, E, t) = F(l, b) \cdot G(E) \cdot H(t)\end{align}
#
# To use full 3D models, ie :math:`f(l, b, E) = F(l, b, E) \cdot \ G(E)`,
# you have to implement your own custom
# `~gammapy.modeling.models.SpatialModel`. Note that it is still necessary to multiply by a
# `~gammapy.modeling.models.SpectralModel`, :math:`G(E)` to be dimensionally consistent.
#
# In this example, we create Gaussian Spatial Model with the extension
# varying with energy. For simplicity, we assume a linear dependency on
# energy and parameterize this by specifying the extension at 2 energies.
# You can add more complex dependencies, probably motivated by physical
# models.
#

from astropy.coordinates import angular_separation
from gammapy.modeling.models import SpatialModel


class MyCustomGaussianModel(SpatialModel):
    """My custom Energy Dependent Gaussian model.

    Parameters
    ----------
    lon_0, lat_0 : `~astropy.coordinates.Angle`
        Center position
    sigma_1TeV : `~astropy.coordinates.Angle`
        Width of the Gaussian at 1 TeV
    sigma_10TeV : `~astropy.coordinates.Angle`
        Width of the Gaussian at 10 TeV

    """

    tag = "MyCustomGaussianModel"
    is_energy_dependent = True
    lon_0 = Parameter("lon_0", "0 deg")
    lat_0 = Parameter("lat_0", "0 deg", min=-90, max=90)

    sigma_1TeV = Parameter("sigma_1TeV", "2.0 deg", min=0)
    sigma_10TeV = Parameter("sigma_10TeV", "0.2 deg", min=0)

    @staticmethod
    def evaluate(lon, lat, energy, lon_0, lat_0, sigma_1TeV, sigma_10TeV):
        sep = angular_separation(lon, lat, lon_0, lat_0)

        # Compute sigma for the given energy using linear interpolation in log energy
        sigma_nodes = u.Quantity([sigma_1TeV, sigma_10TeV])
        energy_nodes = [1, 10] * u.TeV
        log_s = np.log(sigma_nodes.to("deg").value)
        log_en = np.log(energy_nodes.to("TeV").value)
        log_e = np.log(energy.to("TeV").value)
        sigma = np.exp(np.interp(log_e, log_en, log_s)) * u.deg

        exponent = -0.5 * (sep / sigma) ** 2
        norm = 1 / (2 * np.pi * sigma**2)
        return norm * np.exp(exponent)


######################################################################
# Serialisation of this model can be achieved as explained in the previous
# section. You can now use it as standard `~gammapy.modeling.models.SpatialModel` in your
# analysis. Note that this is still a `~gammapy.modeling.models.SpatialModel` and not a
# `~gammapy.modeling.models.SkyModel`, so it needs to be multiplied by a
# `~gammapy.modeling.models.SpectralModel` as before.
#

spatial_model = MyCustomGaussianModel()
spectral_model = PowerLawSpectralModel()
sky_model = SkyModel(spatial_model=spatial_model, spectral_model=spectral_model)

print(spatial_model.evaluation_radius)


######################################################################
# To visualise it, we evaluate it on a 3D geom.
#

energy_axis = MapAxis.from_energy_bounds(
    energy_min=0.1 * u.TeV, energy_max=10.0 * u.TeV, nbin=3, name="energy_true"
)
geom = WcsGeom.create(skydir=(0, 0), width=5.0 * u.deg, binsz=0.1, axes=[energy_axis])

spatial_model.plot_grid(geom=geom, add_cbar=True, figsize=(14, 3))

plt.show()

######################################################################
# For computational purposes, it is useful to specify a
# ``evaluation_radius`` for `~gammapy.modeling.models.SpatialModel` - this gives a size on which
# to compute the model. Though optional, it is highly recommended for
# Custom Spatial Models. This can be done, for ex, by defining the
# following function inside the above class:
#


@property
def evaluation_radius(self):
    """Evaluation radius (`~astropy.coordinates.Angle`)."""
    return 5 * np.max([self.sigma_1TeV.value, self.sigma_10TeV.value]) * u.deg



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Gammapy documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default.
#
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('..'))


import datetime
import sys
import os
import gammapy

# Get configuration information from setup.cfg
from configparser import ConfigParser

# Sphinx-gallery config
from sphinx_gallery.sorting import ExplicitOrder

# Load utils docs functions
from gammapy.utils.docs import SubstitutionCodeBlock, DynamicPRLinkTransform, gammapy_sphinx_ext_activate

# flake8: noqa

# Add our custom directives to Sphinx
def setup(app):
    """
    Add the custom directives to Sphinx.
    """
    app.add_config_value("substitutions", [], "html")
    app.add_directive("substitution-code-block", SubstitutionCodeBlock)
    app.add_post_transform(DynamicPRLinkTransform)


conf = ConfigParser()
conf.read([os.path.join(os.path.dirname(__file__), "..", "setup.cfg")])
setup_cfg = dict(conf.items("metadata"))

sys.path.insert(0, os.path.dirname(__file__))

linkcheck_anchors_ignore = []
linkcheck_ignore = [
    "http://gamma-sky.net/#",
    "https://forge.in2p3.fr/projects/data-challenge-1-dc-1/wiki",
    "https://indico.cta-observatory.org/event/2070/",
    "https://groups.google.com/forum/#!forum/astropy-dev",
    "https://data.hawc-observatory.org/datasets/3hwc-survey/index.php",
    "https://gammapy.slack.com/join/signup#/domain-signup",
    # invalid certificates
    "https://www.hawc-observatory.org/",
    "https://data.hawc-observatory.org/datasets/crab_events_pass4/index.php",
    "https://ipython.org",
    "https://jupyter.org",
    # private pages
    "https://cchesswiki.in2p3.fr/*",
    "https://github.com/VERITAS-Observatory/VEGAS",
]

# the buttons link to html pages which are auto-generated...
linkcheck_exclude_documents = [r"getting-started/.*"]

# -- General configuration ----------------------------------------------------

# By default, highlight as Python 3.
highlight_language = "python3"

# Matplotlib directive sets whether to show a link to the source in HTML
plot_html_show_source_link = False

# If true, figures, tables and code-blocks are automatically numbered if they have a caption
numfig = False

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# Allow to add the canonical html flag on all the pages
# This allows you to indicate to the web search engine that the pages for the stable version are privileged
html_baseurl = 'https://docs.gammapy.org/stable/'

# The reST default role (used for this markup: `text`) to use for all
# documents. Set to the "smart" one.
default_role = 'obj'

# Add any Sphinx extension module names here, as strings.
extensions = [
    # Order for sphinx_automodapi is important
    "sphinx.ext.autosummary",
    "sphinx_automodapi.automodapi", # This should come after autosummary
    "sphinx_automodapi.smart_resolver",
    "sphinx_click.ext",
    'sphinx_copybutton',
    "sphinx_design",
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    'sphinx.ext.viewcode',
    # Allows for mapping to other documentation projects
    "sphinx.ext.intersphinx",
    # Allows for Numpy docstring format
    "numpydoc",
    # Needed for the plot:: functionality in rst
    "matplotlib.sphinxext.plot_directive",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = [
    "**.ipynb_checkpoints",
    "user-guide/model-gallery/*/*.ipynb",
    "user-guide/model-gallery/*/*.md5",
    "user-guide/model-gallery/*/*.py",
    "_build",
]

# Define intersphinx_mapping
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
	"matplotlib": ("https://matplotlib.org/", None),
	"astropy": ("https://docs.astropy.org/en/stable/", None),
	"regions": ("https://astropy-regions.readthedocs.io/en/latest/", None),
	"reproject": ("https://reproject.readthedocs.io/en/latest/", None),
	"naima": ("https://naima.readthedocs.io/en/latest/", None),
	"gadf": ("https://gamma-astro-data-formats.readthedocs.io/en/latest/", None),
	"iminuit": ("https://iminuit.readthedocs.io/en/latest/", None),
	"pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
	}

# -- Options for autosummary/autodoc output ------------------------------------
# Enable generation of stub files
autosummary_generate = True

# Document inherited members
automodsumm_inherited_members = True

# Include class and __init__ docstrings
autoclass_content = "both"

# Directory for API docs
automodapi_toctreedirnm = 'api'

# Suppress member summaries
numpydoc_show_class_members = False

# autodoc options to make the AnalysisConfig look nice
autodoc_class_signature = "separated"
autodoc_typehints = "description"

# Ensures that when users click the "Copy" button, only the actual code is copied,
# excluding interactive prompts and indentation markers
# https://sphinx-copybutton.readthedocs.io/en/latest/use.html#using-regexp-prompt-identifiers
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog = """
.. |Table| replace:: :class:`~astropy.table.Table`
.. |QTable| replace:: :class:`~astropy.table.QTable`
.. |BaseFrame| replace:: :class:`~astropy.coordinates.BaseCoordinateFrame`
.. |SkyCoord| replace:: :class:`~astropy.coordinates.SkyCoord`
"""

# This is added to keep the links to PRs in release notes
changelog_links_docpattern = [".*changelog.*", "whatsnew/.*", "release-notes/.*"]

# -- Project information -------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = setup_cfg["name"]
author = setup_cfg["author"]
copyright = "{}, {}".format(datetime.datetime.now().year, setup_cfg["author"])


version = str(gammapy.__version__)
release = "X.Y.Z" if "dev" in version else version
switch_version = "dev" if "dev" in version else release

substitutions = [
    ("|release|", release),
]
# -- Options for HTML output ---------------------------------------------------

html_theme = "pydata_sphinx_theme"

# Static files to copy after template files
html_static_path = ["_static"]
html_css_files = ["custom.css"]
html_js_files = ["matomo.js"]
templates_path = ["_templates"]


# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_logo = os.path.join(html_static_path[0], "gammapy_logo.png")
html_favicon = os.path.join(html_static_path[0], "gammapy_logo.ico")

# Custom sidebar templates, maps document names to template names.
html_sidebars = {
    "search": ["search-field.html"],
    "navigation": ["sidebar-nav-bs.html"],
}

# If not "", a "Last updated on:" timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = ""

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "{} v{}".format(project, release)

# Output file base name for HTML help builder.
htmlhelp_basename = f"{project}doc"

# Default for the configuration can be found here
# https://github.com/pydata/pydata-sphinx-theme/blob/main/src/pydata_sphinx_theme/theme/pydata_sphinx_theme/theme.conf
html_theme_options = {
    "header_links_before_dropdown": 6,
    "show_toc_level": 2,
    # False = don't show "Previous" and "Next" buttons at the bottom of each page
    "show_prev_next": False,
    # links in menu
    "icon_links": [
        {
            "name": "Github",
            "url": "https://github.com/gammapy/gammapy",
            "icon": "fab fa-github-square",
        },
        {
            "name": "Slack",
            "url": "https://gammapy.slack.com/",
            "icon": "fab fa-slack",
        },
    ],
    "switcher": {
        "json_url": "https://docs.gammapy.org/stable/switcher.json",
        "version_match": switch_version,
    },
    "navbar_end": ["version-switcher", "theme-switcher", "navbar-icon-links"],
    "navigation_with_keys": True,
    # footers
    "footer_start": ["copyright","custom-footer.html"],
    "footer_center": ["last-updated"],
    "footer_end": ["sphinx-version", "theme-version"]
}

gammapy_sphinx_ext_activate()

# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    ("index", f"{project}.tex", f"{project} Documentation", author, "manual")
]

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", project.lower(), f"{project} Documentation", [author], 1)]


# -- Other options -------------------------------------------------------------

github_issues_url = "https://github.com/gammapy/gammapy/issues/"

# In `about.rst` and `references.rst` we are giving lists of citations
# (e.g. papers using Gammapy) that partly aren't referenced from anywhere
# in the Gammapy docs. This is normal, but Sphinx emits a warning.
# The following config option suppresses the warning.
# http://www.sphinx-doc.org/en/stable/rest.html#citations
# http://www.sphinx-doc.org/en/stable/config.html#confval-suppress_warnings
suppress_warnings = ["ref.citation"]

branch = "main" if switch_version == "dev" else f"v{switch_version}"

binder_config = {
    # Required keys
    "org": "gammapy",
    "repo": "gammapy-webpage",
    "branch": branch,  # Can be any branch, tag, or commit hash. Use a branch that hosts your docs.
    "binderhub_url": "https://mybinder.org",  # Any URL of a binderhub deployment. Must be full URL (e.g. https://mybinder.org).
    "dependencies": "./binder/requirements.txt",
    "notebooks_dir": f"notebooks/{switch_version}",
    "use_jupyter_lab": True,
}

sphinx_gallery_conf = {
    # Remove the sphinx comments i.e. sphinx_gallery_thumbnail_number in tutorials
    "remove_config_comments": True,
    "examples_dirs": [
        "../examples/models",
        "../examples/tutorials",
    ],  # path to your example scripts
    "gallery_dirs": [
        "user-guide/model-gallery",
        "tutorials",
    ],  # path to where to save gallery generated output
    "subsection_order": ExplicitOrder(
        [
            "../examples/models/spatial",
            "../examples/models/spectral",
            "../examples/models/temporal",
            "../examples/tutorials/starting",
            "../examples/tutorials/data",
            "../examples/tutorials/model-gallery",
            "../examples/tutorials/details",
            "../examples/tutorials/analysis-1d",
            "../examples/tutorials/analysis-2d",
            "../examples/tutorials/analysis-3d",
            "../examples/tutorials/analysis-time",
            "../examples/tutorials/astrophysics",
            "../examples/tutorials/scripts",
        ]
    ),
    "binder": binder_config,
    "backreferences_dir": "gen_modules/backreferences",
    "doc_module": ("gammapy",),
    "exclude_implicit_doc": {},
    "filename_pattern": r"\.py",
    "reset_modules": ("matplotlib",),
    "within_subsection_order": "sphinxext.TutorialExplicitOrder",
    "download_all_examples": True,
    "capture_repr": ("_repr_html_", "__repr__"),
    # Show sidebar dropdowns for menu
    "nested_sections": True,
    "min_reported_time": 10,
    "show_memory": False,
    "line_numbers": False,
    "reference_url": {
        # The module you locally document uses None
        "gammapy": None,
    },
}

html_context = {
    "default_mode": "light",
}




"""Plot significance image with HESS and MILAGRO colormap."""

from astropy.visualization import LinearStretch
from astropy.visualization.mpl_normalize import ImageNormalize
import matplotlib.pyplot as plt
from gammapy.maps import Map
from gammapy.visualization import colormap_hess, colormap_milagro

filename = "$GAMMAPY_DATA/tests/unbundled/poisson_stats_image/expected_ts_0.000.fits.gz"
image = Map.read(filename, hdu="SQRT_TS")

# Plot with the HESS and Milagro colormap
vmin, vmax, vtransition = -5, 15, 5
fig = plt.figure(figsize=(15.5, 6))

normalize = ImageNormalize(vmin=vmin, vmax=vmax, stretch=LinearStretch())
transition = normalize(vtransition)

ax = fig.add_subplot(121, projection=image.geom.wcs)
cmap = colormap_hess(transition=transition)
image.plot(ax=ax, cmap=cmap, norm=normalize, add_cbar=True)
plt.title("HESS-style colormap")

ax = fig.add_subplot(122, projection=image.geom.wcs)
cmap = colormap_milagro(transition=transition)
image.plot(ax=ax, cmap=cmap, norm=normalize, add_cbar=True)
plt.title("MILAGRO-style colormap")

plt.tight_layout()
plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from typing import ClassVar, Literal, Optional
from pydantic import Field
from gammapy.utils.fits import earth_location_from_dict
from gammapy.utils.metadata import (
    METADATA_FITS_KEYS,
    CreatorMetaData,
    MetaData,
    ObsInfoMetaData,
    PointingInfoMetaData,
    TargetMetaData,
    TimeInfoMetaData,
)
from gammapy.utils.time import time_ref_from_dict, time_ref_to_dict
from gammapy.utils.types import EarthLocationType, TimeType

__all__ = ["ObservationMetaData", "GTIMetaData", "EventListMetaData"]

OBSERVATION_METADATA_FITS_KEYS = {
    "location": {
        "input": lambda v: earth_location_from_dict(v),
        "output": lambda v: {
            "GEOLON": v.lon.deg,
            "GEOLAT": v.lat.deg,
            "ALTITUDE": v.height.to_value("m"),
        },
    },
    "deadtime_fraction": {
        "input": lambda v: 1 - v["DEADC"],
        "output": lambda v: {"DEADC": 1 - v},
    },
}

METADATA_FITS_KEYS["observation"] = OBSERVATION_METADATA_FITS_KEYS


EVENTLIST_METADATA_FITS_KEYS = {
    "event_class": "EV_CLASS",
}

METADATA_FITS_KEYS["eventlist"] = EVENTLIST_METADATA_FITS_KEYS
METADATA_FITS_KEYS["GTI"] = {}


class ObservationMetaData(MetaData):
    """Metadata containing information about the Observation.

    Parameters
    ----------
    obs_info : `~gammapy.utils.ObsInfoMetaData`
        The general observation information.
    pointing : `~gammapy.utils.PointingInfoMetaData`
        The pointing metadata.
    target : `~gammapy.utils.TargetMetaData`
        The target metadata.
    creation : `~gammapy.utils.CreatorMetaData`
        The creation metadata.
    location : `~astropy.coordinates.EarthLocation` or str, optional
        The observatory location.
    deadtime_fraction : float
        The observation deadtime fraction. Default is 0.
    optional : dict, optional
        Additional optional metadata.
    """

    _tag: ClassVar[Literal["observation"]] = "observation"
    obs_info: Optional[ObsInfoMetaData] = None
    pointing: Optional[PointingInfoMetaData] = None
    target: Optional[TargetMetaData] = None
    location: Optional[EarthLocationType] = None
    deadtime_fraction: float = Field(0.0, ge=0, le=1.0)
    time_info: Optional[TimeInfoMetaData] = None
    creation: Optional[CreatorMetaData] = None
    optional: Optional[dict] = None

    @classmethod
    def from_header(cls, header, format="gadf"):
        """Create and fill the observation metadata from the event list metadata.

        Parameters
        ----------
        header : dict
            Input FITS header.
        format : str
            The header data format. Default is gadf.
        """
        meta = super().from_header(header, format)

        meta.creation = CreatorMetaData()
        # Include additional gadf keywords not specified as ObservationMetaData attributes
        optional_keywords = [
            "OBSERVER",
            "EV_CLASS",
            "TELAPSE",
            "TELLIST",
            "N_TELS",
            "TASSIGN",
            "DST_VER",
            "ANA_VER",
            "CAL_VER",
            "CONV_DEP",
            "CONV_RA",
            "CONV_DEC",
            "TRGRATE",
            "ZTRGRATE",
            "MUONEFF",
            "BROKPIX",
            "AIRTEMP",
            "PRESSURE",
            "RELHUM",
            "NSBLEVEL",
            "CREATOR",
            "HDUVERS",
        ]
        optional = dict()
        for key in optional_keywords:
            if key in header.keys():
                optional[key] = header[key]
        meta.optional = optional

        return meta


class GTIMetaData(MetaData):
    """Metadata containing information about the GTI.

    Parameters
    ----------
    reference_time : Time, str
        The GTI reference time.
    creation : `~gammapy.utils.metadata.CreatorMetaData`
        The creation metadata.
    """

    _tag: ClassVar[Literal["GTI"]] = "GTI"
    reference_time: Optional[TimeType] = None
    creation: Optional[CreatorMetaData] = None

    def to_header(self, format="gadf"):
        result = super().to_header(format)
        if self.reference_time is not None:
            result.update(
                time_ref_to_dict(self.reference_time, scale=self.reference_time.scale)
            )
        return result

    @classmethod
    def from_header(cls, header, format="gadf"):
        meta = super().from_header(header, format)
        try:
            meta.reference_time = time_ref_from_dict(header)
        except KeyError:
            meta.reference_time = None

        return meta


class EventListMetaData(MetaData):
    """
    Metadata containing information about the EventList.

    Parameters
    ----------
    event_class : str
        The event class metadata.
    creation : `~gammapy.utils.metadata.CreatorMetaData`
        The creation metadata.
    """

    _tag: ClassVar[Literal["EventList"]] = "eventlist"
    event_class: Optional[str] = None
    creation: Optional[CreatorMetaData] = None
    optional: Optional[dict] = None

    @classmethod
    def from_header(cls, header, format="gadf"):
        meta = super().from_header(header, format)

        # Include additional gadf keywords
        optional_keywords = [
            "DST_VER",
            "ANA_VER",
            "CAL_VER",
        ]
        optional = dict()
        for key in optional_keywords:
            if key in header.keys():
                optional[key] = header[key]
        meta.optional = optional

        return meta



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from itertools import product
import numpy as np
from scipy.ndimage import label as ndi_label
from astropy import units as u
from astropy.io import fits
from astropy.nddata import block_reduce
from astropy.table import Table
from astropy.visualization import quantity_support
import matplotlib.pyplot as plt
from gammapy.maps.axes import UNIT_STRING_FORMAT
from gammapy.utils.interpolation import ScaledRegularGridInterpolator, StatProfileScale
from gammapy.utils.scripts import make_path
from ..axes import MapAxes
from ..core import Map
from ..geom import pix_tuple_to_idx
from ..region import RegionGeom
from ..utils import INVALID_INDEX

__all__ = ["RegionNDMap"]


class RegionNDMap(Map):
    """N-dimensional region map.

    A `~RegionNDMap` owns a `~RegionGeom` instance as well as a data array
    containing the values associated to that region in the sky along the non-spatial
    axis, usually an energy axis. The spatial dimensions of a `~RegionNDMap`
    are reduced to a single spatial bin with an arbitrary shape,
    and any extra dimensions are described by an arbitrary number of non-spatial axes.

    Parameters
    ----------
    geom : `~gammapy.maps.RegionGeom`
        Region geometry object.
    data : `~numpy.ndarray`
        Data array. If None then an empty array will be allocated.
    dtype : str, optional
        Data type. Default is "float32".
    meta : `dict`, optional
        Dictionary to store metadata.
        Default is None.
    unit : str or `~astropy.units.Unit`, optional
        The map unit.
        Default is "".

    Notes
    -----
    - For further explanation and examples, see :doc:`/user-guide/maps/regionmap`
    """

    def __init__(self, geom, data=None, dtype="float32", meta=None, unit=""):
        if data is None:
            data = np.zeros(geom.data_shape, dtype=dtype)

        if meta is None:
            meta = {}

        self._geom = geom
        self.data = data
        self.meta = meta
        self._unit = u.Unit(unit)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        """Set data.

        Parameters
        ----------
        value : array-like
            Data array.
        """
        if np.isscalar(value):
            value = value * np.ones(self.geom.data_shape, dtype=type(value))

        if isinstance(value, u.Quantity):
            raise TypeError("Map data must be a Numpy array. Set unit separately")

        input_shape = value.shape
        if len(self.geom.data_shape) == 2 + len(value.shape):
            if self.geom.data_shape[: len(value.shape)] == value.shape:
                value = np.expand_dims(value, (-2, -1))

        if self.geom.data_shape != value.shape:
            raise ValueError(
                f"Input shape {input_shape} is not compatible with shape from geometry {self.geom.data_shape}"
            )

        self._data = value

    def plot(self, ax=None, axis_name=None, **kwargs):
        """Plot the data contained in region map along the non-spatial axis.

        Parameters
        ----------
        ax : `~matplotlib.pyplot.Axis`, optional
            Axis used for plotting.
            Default is None.
        axis_name : str, optional
            Which axis to plot on the x-axis. Extra axes will be plotted as
            additional lines. Default is None.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.errorbar`.

        Returns
        -------
        ax : `~matplotlib.pyplot.Axis`
            Axis used for plotting.
        """
        ax = ax or plt.gca()

        if axis_name is None:
            if self.geom.axes.is_unidimensional:
                axis_name = self.geom.axes.primary_axis.name
            else:
                raise ValueError(
                    "Plotting a region map with multiple extra axes requires "
                    "specifying the 'axis_name' keyword."
                )

        axis = self.geom.axes[axis_name]

        kwargs.setdefault("marker", "o")
        kwargs.setdefault("markersize", 4)
        kwargs.setdefault("ls", "None")
        kwargs.setdefault("xerr", axis.as_plot_xerr)

        yerr_nd, yerr = kwargs.pop("yerr", None), None
        uplims_nd, uplims = kwargs.pop("uplims", None), None
        label_default = kwargs.pop("label", None)

        labels = product(
            *[ax.as_plot_labels for ax in self.geom.axes if ax.name != axis.name]
        )

        for label_axis, (idx, quantity) in zip(
            labels, self.iter_by_axis_data(axis_name=axis.name)
        ):
            if isinstance(yerr_nd, tuple):
                yerr = yerr_nd[0][idx], yerr_nd[1][idx]
            elif isinstance(yerr_nd, np.ndarray):
                yerr = yerr_nd[idx]

            if uplims_nd is not None:
                uplims = uplims_nd[idx]

            label = " ".join(label_axis) if label_default is None else label_default

            with quantity_support():
                ax.errorbar(
                    x=axis.as_plot_center,
                    y=quantity,
                    yerr=yerr,
                    uplims=uplims,
                    label=label,
                    **kwargs,
                )
        axis.format_plot_xaxis(ax=ax)

        if "energy" in axis_name:
            ax.set_yscale("log", nonpositive="clip")

        return ax

    def plot_hist(self, ax=None, **kwargs):
        """Plot as histogram.

        Parameters
        ----------
        ax : `~matplotlib.axis`, optional
            Axis instance to be used for the plot.
            Default is None.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.hist`.

        Returns
        -------
        ax : `~matplotlib.pyplot.Axis`
            Axis used for plotting.
        """
        ax = plt.gca() if ax is None else ax

        kwargs.setdefault("histtype", "step")
        kwargs.setdefault("lw", 1)

        if not self.geom.axes.is_unidimensional:
            raise ValueError("Plotting is only supported for unidimensional maps")

        axis = self.geom.axes[0]

        with quantity_support():
            weights = self.data[:, 0, 0]
            ax.hist(
                axis.as_plot_center, bins=axis.as_plot_edges, weights=weights, **kwargs
            )

        if not self.unit.is_unity():
            ax.set_ylabel(f"Data [{self.unit.to_string(UNIT_STRING_FORMAT)}]")

        axis.format_plot_xaxis(ax=ax)
        ax.set_yscale("log")
        return ax

    def plot_interactive(self):
        raise NotImplementedError(
            "Interactive plotting currently not support for RegionNDMap"
        )

    def plot_region(self, ax=None, **kwargs):
        """Plot region.

        Parameters
        ----------
        ax : `~astropy.visualization.WCSAxes`, optional
            Axes to plot on. If no axes are given,
            the region is shown using the minimal
            equivalent WCS geometry. Default is None.
        **kwargs : dict
            Keyword arguments forwarded to `~regions.PixelRegion.as_artist`.
        """
        ax = self.geom.plot_region(ax, **kwargs)
        return ax

    def plot_mask(self, ax=None, **kwargs):
        """Plot the mask as a shaded area in a xmin-xmax range.

        Parameters
        ----------
        ax : `~matplotlib.axis`, optional
            Axis instance to be used for the plot.
            Default is None.
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.pyplot.axvspan`.

        Returns
        -------
        ax : `~matplotlib.pyplot.Axis`
            Axis used for plotting.
        """
        if not self.is_mask:
            raise ValueError("This is not a mask and cannot be plotted")

        kwargs.setdefault("color", "k")
        kwargs.setdefault("alpha", 0.05)
        kwargs.setdefault("label", "mask")

        ax = plt.gca() if ax is None else ax

        edges = self.geom.axes["energy"].edges.reshape((-1, 1, 1))

        labels, nlabels = ndi_label(self.data)

        for idx in range(1, nlabels + 1):
            mask = labels == idx
            xmin = edges[:-1][mask].min().value
            xmax = edges[1:][mask].max().value
            ax.axvspan(xmin, xmax, **kwargs)

        return ax

    @classmethod
    def create(
        cls,
        region,
        axes=None,
        dtype="float32",
        meta=None,
        unit="",
        wcs=None,
        binsz_wcs="0.1 deg",
        data=None,
    ):
        """Create an empty region map object.

        Parameters
        ----------
        region : str or `~regions.SkyRegion`
            Region specification.
        axes : list of `MapAxis`, optional
            Non-spatial axes. Default is None.
        dtype : str, optional
            Data type. Default is 'float32'.
        unit : str or `~astropy.units.Unit`, optional
            Data unit. Default is "".
        meta : `dict`, optional
            Dictionary to store metadata.
            Default is None.
        wcs : `~astropy.wcs.WCS`, optional
            WCS projection to use for local projections of the region.
            Default is None.
        binsz_wcs: `~astropy.units.Quantity` or str, optional
            Bin size used for the default WCS, if ``wcs=None``.
            Default is "0.1 deg".
        data : `~numpy.ndarray`, optional
            Data array. Default is None.

        Returns
        -------
        map : `RegionNDMap`
            Region map.
        """
        geom = RegionGeom.create(region=region, axes=axes, wcs=wcs, binsz_wcs=binsz_wcs)
        return cls(geom=geom, dtype=dtype, unit=unit, meta=meta, data=data)

    def downsample(self, factor, preserve_counts=True, axis_name=None, weights=None):
        """Downsample the non-spatial dimension by a given factor.

        By default, the first axes is downsampled.

        Parameters
        ----------
        factor : int
            Downsampling factor.
        preserve_counts : bool, optional
            Preserve the integral over each bin.  This should be true
            if the map is an integral quantity (e.g. counts) and false if
            the map is a differential quantity (e.g. intensity).
            Default is True.
        axis_name : str, optional
            Which axis to downsample. Default is "energy".
        weights : `RegionNDMap`, optional
            Contains the weights to apply to the axis to reduce. Default
            weights of one.

        Returns
        -------
        map : `RegionNDMap`
            Downsampled region map.
        """
        if axis_name is None:
            axis_name = self.geom.axes[0].name

        geom = self.geom.downsample(factor=factor, axis_name=axis_name)

        block_size = [1] * self.data.ndim
        idx = self.geom.axes.index_data(axis_name)
        block_size[idx] = factor

        if weights is None:
            weights = 1
        else:
            weights = weights.data

        func = np.nansum if preserve_counts else np.nanmean
        if self.is_mask:
            func = np.all
        data = block_reduce(self.data * weights, tuple(block_size), func=func)

        return self._init_copy(geom=geom, data=data)

    def upsample(self, factor, order=0, preserve_counts=True, axis_name=None):
        """Upsample the non-spatial dimension by a given factor.

        By default, the first axes is upsampled.

        Parameters
        ----------
        factor : int
            Upsampling factor.
        order : int, optional
            Order of the interpolation used for upsampling.
            Default is 0.
        preserve_counts : bool, optional
            Preserve the integral over each bin.  This should be true
            if the RegionNDMap is an integral quantity (e.g. counts) and false if
            the RegionNDMap is a differential quantity (e.g. intensity).
            Default is True.
        axis_name : str, optional
            Which axis to upsample. Default is "energy".

        Returns
        -------
        map : `RegionNDMap`
            Upsampled region map.
        """
        if axis_name is None:
            axis_name = self.geom.axes[0].name

        geom = self.geom.upsample(factor=factor, axis_name=axis_name)
        data = self.interp_by_coord(geom.get_coord())

        if preserve_counts:
            data /= factor

        return self._init_copy(geom=geom, data=data)

    def iter_by_axis_data(self, axis_name):
        """Iterate data by axis.

        Parameters
        ----------
        axis_name : str
            Axis name.

        Returns
        -------
        idx, data : tuple, `~astropy.units.Quantity`
            Data and index.
        """
        idx_axis = self.geom.axes.index_data(axis_name)
        shape = list(self.data.shape)
        shape[idx_axis] = 1

        for idx in np.ndindex(*shape):
            idx = list(idx)
            idx[idx_axis] = slice(None)
            yield tuple(idx), self.quantity[tuple(idx)]

    def _resample_by_idx(self, idx, weights=None, preserve_counts=False, smooth=False):
        # inherited docstring
        # TODO: too complex, simplify!
        idx = pix_tuple_to_idx(idx)

        msk = np.all(np.stack([t != INVALID_INDEX.int for t in idx]), axis=0)
        idx = [t[msk] for t in idx]

        if weights is not None:
            if isinstance(weights, u.Quantity):
                weights = weights.to_value(self.unit)
            weights = weights[msk]

        idx = np.ravel_multi_index(idx, self.data.T.shape)
        idx, idx_inv = np.unique(idx, return_inverse=True)
        weights = np.bincount(idx_inv, weights=weights).astype(self.data.dtype)
        if not preserve_counts:
            weights /= np.bincount(idx_inv).astype(self.data.dtype)
        self.data.T.flat[idx] += weights

    def fill_by_idx(self, idx, weights=None):
        return self._resample_by_idx(idx, weights=weights, preserve_counts=True)

    def get_by_idx(self, idxs):
        # inherited docstring
        return self.data[idxs[::-1]]

    def interp_by_coord(self, coords, **kwargs):
        """Interpolate map values at the given map coordinates.

        Parameters
        ----------
        coords : tuple, dict or `~gammapy.maps.MapCoord`
            Coordinate arrays for each dimension of the map.  Tuple
            should be ordered as (lon, lat, x_0, ..., x_n) where x_i
            are coordinates for non-spatial dimensions of the map.
            "lon" and "lat" are optional and will be taken at the center
            of the region by default.
        method : {"linear", "nearest"}
            Method to interpolate data values. Default is "linear".
        fill_value : None or float value
            The value to use for points outside the interpolation domain.
            If None, values outside the domain are extrapolated.
        values_scale : {"lin", "log", "sqrt"}
            Optional value scaling.

        Returns
        -------
        vals : `~numpy.ndarray`
            Interpolated pixel values.
        """
        pix = self.geom.coord_to_pix(coords=coords)
        return self.interp_by_pix(pix, **kwargs)

    def interp_by_pix(self, pix, **kwargs):
        # inherited docstring
        grid_pix = [np.arange(n, dtype=float) for n in self.data.shape[::-1]]

        if np.any(np.isfinite(self.data)):
            data = self.data.copy().T
            data[~np.isfinite(data)] = 0.0
        else:
            data = self.data.T

        scale = kwargs.get("values_scale", "lin")

        if scale == "stat-profile":
            axis = 2 + self.geom.axes.index("norm")
            kwargs["values_scale"] = StatProfileScale(axis=axis)

        fn = ScaledRegularGridInterpolator(grid_pix, data, **kwargs)
        return fn(tuple(pix), clip=False)

    def set_by_idx(self, idx, value):
        # inherited docstring
        self.data[idx[::-1]] = value

    @classmethod
    def read(cls, filename, format="gadf", ogip_column=None, hdu=None, checksum=False):
        """Read from file.

        Parameters
        ----------
        filename : `pathlib.Path` or str
            Filename.
        format : {"gadf", "ogip", "ogip-arf"}
            Which format to use. Default is "gadf".
        ogip_column : {None, "COUNTS", "QUALITY", "BACKSCAL"}, optional
            If format 'ogip' is chosen which table hdu column to read.
            Default is None.
        hdu : str, optional
            Name or index of the HDU with the map data.
            Default is None.
        checksum : bool
            If True checks both DATASUM and CHECKSUM cards in the file headers. Default is False.

        Returns
        -------
        region_map : `RegionNDMap`
            Region map.
        """
        filename = make_path(filename)
        with fits.open(filename, memmap=False, checksum=checksum) as hdulist:
            return cls.from_hdulist(
                hdulist, format=format, ogip_column=ogip_column, hdu=hdu
            )

    def write(
        self,
        filename,
        overwrite=False,
        format="gadf",
        hdu="SKYMAP",
        checksum=False,
    ):
        """Write map to file.

        Parameters
        ----------
        filename : `pathlib.Path` or str
            Filename.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        format : {"gadf", "ogip", "ogip-sherpa", "ogip-arf", "ogip-arf-sherpa"}
            Which format to use. Default is "gadf".
        hdu : str, optional
            Name of the HDU. Default is "SKYMAP".
        checksum : bool, optional
            When True adds both DATASUM and CHECKSUM cards to the headers written to the file.
            Default is False.
        """
        filename = make_path(filename)
        self.to_hdulist(format=format, hdu=hdu).writeto(
            filename, overwrite=overwrite, checksum=checksum
        )

    def to_hdulist(self, format="gadf", hdu="SKYMAP", hdu_bands=None, hdu_region=None):
        """Convert to `~astropy.io.fits.HDUList`.

        Parameters
        ----------
        format : {"gadf", "ogip", "ogip-sherpa", "ogip-arf", "ogip-arf-sherpa"}
            Format specification. Default is "gadf".
        hdu : str, optional
            Name of the HDU with the map data, used for "gadf" format.
            Default is "SKYMAP".
        hdu_bands : str, optional
            Name or index of the HDU with the BANDS table, used for "gadf" format.
            Default is None.
        hdu_region : str, optional
            Name or index of the HDU with the region table.
            Default is None.

        Returns
        -------
        hdulist : `~astropy.fits.HDUList`
            HDU list.
        """
        hdulist = fits.HDUList()
        table = self.to_table(format=format)

        if hdu_bands is None:
            hdu_bands = f"{hdu.upper()}_BANDS"
        if hdu_region is None:
            hdu_region = f"{hdu.upper()}_REGION"

        if format in ["ogip", "ogip-sherpa", "ogip-arf", "ogip-arf-sherpa"]:
            hdulist.append(fits.BinTableHDU(table))
        elif format == "gadf":
            table.meta.update(self.geom.axes.to_header())
            hdulist.append(fits.BinTableHDU(table, name=hdu))
        else:
            raise ValueError(f"Unsupported format '{format}'")

        if format in ["ogip", "ogip-sherpa", "gadf"]:
            hdulist_geom = self.geom.to_hdulist(
                format=format, hdu_bands=hdu_bands, hdu_region=hdu_region
            )
            hdulist.extend(hdulist_geom[1:])

        return hdulist

    @classmethod
    def from_table(cls, table, format="", colname=None):
        """Create region map from table.

        Parameters
        ----------
        table : `~astropy.table.Table`
            Table with input data.
        format : {"gadf-sed", "lightcurve", "profile"}
            Format to use.
        colname : str
            Column name to take the data from.

        Returns
        -------
        region_map : `RegionNDMap`
            Region map.
        """
        if format == "gadf-sed":
            if colname is None:
                raise ValueError("Column name required")

            axes = MapAxes.from_table(table=table, format=format)

            if colname == "stat_scan":
                names = ["norm", "energy"]
            # TODO: this is not officially supported by GADF...
            elif colname in ["counts", "npred", "npred_excess"]:
                names = ["dataset", "energy"]
            else:
                names = ["energy"]

            axes = axes[names]
            data = table[colname].data
            unit = table[colname].unit or ""
        elif format == "lightcurve":
            axes = MapAxes.from_table(table=table, format=format)

            if colname == "stat_scan":
                names = ["norm", "energy", "time"]
            # TODO: this is not officially supported by GADF...
            elif colname in ["counts", "npred", "npred_excess"]:
                names = ["dataset", "energy", "time"]
            else:
                names = ["energy", "time"]

            axes = axes[names]
            data = table[colname].data
            unit = table[colname].unit or ""
        elif format == "profile":
            axes = MapAxes.from_table(table=table, format=format)

            if colname == "stat_scan":
                names = ["norm", "energy", "projected-distance"]
            # TODO: this is not officially supported by GADF...
            elif colname in ["counts", "npred", "npred_excess"]:
                names = ["dataset", "energy", "projected-distance"]
            else:
                names = ["energy", "projected-distance"]

            axes = axes[names]
            data = table[colname].data
            unit = table[colname].unit or ""
        else:
            raise ValueError(f"Format not supported {format}")

        geom = RegionGeom.create(region=None, axes=axes)
        data = data.reshape(geom.data_shape)
        return cls(geom=geom, data=data, unit=unit, meta=table.meta, dtype=data.dtype)

    @classmethod
    def from_hdulist(cls, hdulist, format="gadf", ogip_column=None, hdu=None, **kwargs):
        """Create from `~astropy.io.fits.HDUList`.

        Parameters
        ----------
        hdulist : `~astropy.io.fits.HDUList`
            HDU list.
        format : {"gadf", "ogip", "ogip-arf"}
            Format specification. Default is "gadf".
        ogip_column : {"COUNTS", "QUALITY", "BACKSCAL"}, optional
            If format 'ogip' is chosen which table HDU column to read.
            Default is None.
        hdu : str, optional
            Name or index of the HDU with the map data.
            Default is None.

        Returns
        -------
        region_nd_map : `RegionNDMap`
            Region map.
        """
        defaults = {
            "ogip": {"hdu": "SPECTRUM", "column": "COUNTS"},
            "ogip-arf": {"hdu": "SPECRESP", "column": "SPECRESP"},
            "gadf": {"hdu": "SKYMAP", "column": "DATA"},
        }

        if hdu is None:
            hdu = defaults[format]["hdu"]

        if ogip_column is None:
            ogip_column = defaults[format]["column"]

        geom = RegionGeom.from_hdulist(hdulist, format=format, hdu=hdu)

        table = Table.read(hdulist[hdu])
        quantity = table[ogip_column].quantity.reshape(geom.data_shape)

        if ogip_column == "QUALITY":
            data, unit = np.logical_not(quantity.value.astype(bool)), ""
        else:
            data, unit = quantity.value, quantity.unit

        return cls(geom=geom, data=data, meta=table.meta, unit=unit, dtype=data.dtype)

    def _pad_spatial(self, *args, **kwargs):
        raise NotImplementedError("Spatial padding is not supported by RegionNDMap")

    def crop(self):
        raise NotImplementedError("Crop is not supported by RegionNDMap")

    def stack(self, other, weights=None, nan_to_num=True):
        """Stack other region map into map.

        Parameters
        ----------
        other : `RegionNDMap`
            Other map to stack.
        weights : `RegionNDMap`, optional
            Array to be used as weights. The spatial geometry must be equivalent
            to `other` and additional axes must be broadcastable.
            Default is None.
        nan_to_num: bool, optional
            Non-finite values are replaced by zero if True.
            Default is True.
        """
        data = other.quantity.to_value(self.unit).astype(self.data.dtype)

        # TODO: re-think stacking of regions. Is making the union reasonable?
        # self.geom.union(other.geom)
        if nan_to_num:
            not_finite = ~np.isfinite(data)
            if np.any(not_finite):
                data = data.copy()
                data[not_finite] = 0
        if weights is not None:
            if not other.geom.to_image() == weights.geom.to_image():
                raise ValueError("Incompatible geoms between map and weights")
            data = data * weights.data

        self.data += data

    def to_table(self, format="gadf"):
        """Convert to `~astropy.table.Table`.

        Data format specification: :ref:`gadf:ogip-pha`.

        Parameters
        ----------
        format : {"gadf", "ogip", "ogip-arf", "ogip-arf-sherpa"}
            Format specification. Default is "gadf".

        Returns
        -------
        table : `~astropy.table.Table`
            Table.
        """
        data = np.nan_to_num(self.quantity[:, 0, 0])

        if format == "ogip":
            if len(self.geom.axes) > 1:
                raise ValueError(
                    f"Writing to format '{format}' only supports a "
                    f"single energy axis. Got {self.geom.axes.names}"
                )

            energy_axis = self.geom.axes[0]
            energy_axis.assert_name("energy")
            table = Table()
            table["CHANNEL"] = np.arange(energy_axis.nbin, dtype=np.int16)
            table["COUNTS"] = np.array(data, dtype=np.int32)

            # see https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/node6.html  # noqa: E501
            table.meta = {
                "EXTNAME": "SPECTRUM",
                "telescop": "unknown",
                "instrume": "unknown",
                "filter": "None",
                "exposure": 0,
                "corrfile": "",
                "corrscal": "",
                "ancrfile": "",
                "hduclass": "OGIP",
                "hduclas1": "SPECTRUM",
                "hduvers": "1.2.1",
                "poisserr": True,
                "chantype": "PHA",
                "detchans": energy_axis.nbin,
                "quality": 0,
                "backscal": 0,
                "grouping": 0,
                "areascal": 1,
            }

        elif format in ["ogip-arf", "ogip-arf-sherpa"]:
            if len(self.geom.axes) > 1:
                raise ValueError(
                    f"Writing to format '{format}' only supports a "
                    f"single energy axis. Got {self.geom.axes.names}"
                )

            energy_axis = self.geom.axes[0]
            table = energy_axis.to_table(format=format)
            table.meta = {
                "EXTNAME": "SPECRESP",
                "telescop": "unknown",
                "instrume": "unknown",
                "filter": "None",
                "hduclass": "OGIP",
                "hduclas1": "RESPONSE",
                "hduclas2": "SPECRESP",
                "hduvers": "1.1.0",
            }

            if format == "ogip-arf-sherpa":
                data = data.to("cm2")

            table["SPECRESP"] = data

        elif format == "gadf":
            table = Table()
            data = self.quantity
            table["CHANNEL"] = np.arange(len(data), dtype=np.int16)
            table["DATA"] = data
        else:
            raise ValueError(f"Unsupported format: '{format}'")

        meta = {k: self.meta.get(k, v) for k, v in table.meta.items()}
        table.meta.update(meta)
        return table

    def get_spectrum(self, *args, **kwargs):
        """Return self."""
        return self

    def to_region_nd_map(self, *args, **kwargs):
        """Return self."""
        return self

    def cutout(self, *args, **kwargs):
        """Return self."""
        return self



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import copy
import inspect
import logging
import numpy as np
from gammapy.modeling.models import DatasetModels
from .core import Dataset, Datasets
from .map import MapDataset

log = logging.getLogger(__name__)


__all__ = ["DatasetsActor"]


class DatasetsActor(Datasets):
    """A modified Dataset collection for parallel evaluation using ray actors.

    Support only datasets composed of MapDataset.

    Parameters
    ----------
    datasets : `Datasets`
        Datasets
    """

    def __init__(self, datasets=None):
        from ray import get

        log.warning(
            "Gammapy support for parallelisation with ray is still a prototype and is not fully functional."
        )

        if datasets is not None:
            datasets_list = []
            while datasets:
                d0 = datasets[0]
                if d0.tag != "MapDataset":
                    raise TypeError(
                        f"For now datasets parallel evaluation is only supported for MapDataset, got {d0.tag} instead"
                    )
                if isinstance(d0, MapDatasetActor):
                    datasets_list.append(d0)
                else:
                    datasets_list.append(MapDatasetActor(d0))
                datasets.remove(d0)  # moved to remote so removed from main process
            self._datasets = datasets_list
            self._ray_get = get
            self._covariance = None

        # trigger actors auto_init_wrapper (so overhead so appears on init)
        self.name

    def insert(self, idx, dataset):
        if isinstance(dataset, Dataset):
            if dataset.name in self.names:
                raise (ValueError("Dataset names must be unique"))
            self._datasets.insert(idx, MapDatasetActor(dataset))
        else:
            raise TypeError(f"Invalid type: {type(dataset)!r}")

    def __getattr__(self, name):
        """Get attribute from remote each dataset."""

        def wrapper(*args, **kwargs):
            results = self._ray_get(
                [
                    d._get_remote(name, *args, **kwargs, from_actors=True)
                    for d in self._datasets
                ]
            )
            if "plot" in name:
                results = [res(**kwargs) for res in results]
            for d in self._datasets:
                d._to_update = {}
            return results

        if inspect.ismethod(getattr(self._datasets[0], name)):
            return wrapper
        else:
            return wrapper()

    def stat_sum(self):
        """Compute joint likelihood."""
        results = self._ray_get([d._update_stat_sum_remote() for d in self._datasets])
        return np.sum(results)

    def _to_asimov_datasets(self):
        """Create Asimov datasets from the current models."""
        asimov_datasets = Datasets([d._to_asimov_dataset() for d in self])
        for d in asimov_datasets:
            d._models = self.models
        return asimov_datasets


class RayFrontendMixin(object):
    """Ray mixin for a local class that interact with a remote instance."""

    # TODO: abstract class ?

    @property
    def _remote_attr(self):
        return [
            key
            for key in self._public_attr
            if key not in self._mutable_attr + self._local_attr
        ]

    def __getattr__(self, name):
        """Get attribute from remote."""
        if name in self._remote_attr:
            results = self._ray_get(self._get_remote(name, from_actors=False))
            return results
        else:
            return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if name in self._remote_attr:
            raise AttributeError("can't set attribute")
        else:
            super().__setattr__(name, value)
            if name in self._mutable_attr:
                self._to_update[name] = value
                self._cache[name] = copy.deepcopy(value)

    def _get_remote(self, attr, *args, from_actors=False, **kwargs):
        results = self._actor._get.remote(
            attr, *args, **kwargs, to_update=self._to_update, from_actors=from_actors
        )
        self._to_update = {}
        return results


class MapDatasetActor(RayFrontendMixin):
    """MapDataset for parallel evaluation as a ray actor.

    Parameters
    ----------
    dataset : `MapDataset`
        MapDataset
    """

    _mutable_attr = ["models", "mask_fit"]
    _local_attr = ["name"]  # immutable small enough to keep and acces locally
    _public_attr = [key for key in dir(MapDataset) if not key.startswith("__")]

    def __init__(self, dataset):
        from ray import get, remote

        self._ray_get = get
        self._actor = remote(_MapDatasetActorBackend).remote()
        self._actor._from_dataset.remote(dataset)
        self._name = dataset.name
        self._to_update = {}
        self._cache = {}
        if dataset.models is None:
            self.models = DatasetModels()
        else:
            self.models = dataset.models
        self.mask_fit = dataset.mask_fit
        self._to_update = {}  # models and mask_fit are already ok from actor init

    @property
    def name(self):
        return self._name

    def _update_stat_sum_remote(self):
        self._check_parameters()
        values = self.models.parameters.free_parameters.value
        output = self._actor._update_stat_sum.remote(values, to_update=self._to_update)
        self._to_update = {}
        return output

    def __setattr__(self, name, value):
        if name == "models":
            if value is None:
                value = DatasetModels()
            value = value.select(datasets_names=self.name)
        super().__setattr__(name, value)

    def _get_remote(self, attr, *args, from_actors=False, **kwargs):
        self._check_models()
        results = super()._get_remote(attr, *args, from_actors=False, **kwargs)
        return results

    def _check_models(self):
        if ~np.all(
            self.models.parameters.value == self._cache["models"].parameters.value
        ) or len(self.models.parameters.free_parameters) != len(
            self._cache["models"].parameters.free_parameters
        ):
            self._to_update["models"] = self.models
            self._cache["models"] = self.models.copy()

    def _check_parameters(self):
        if self.models.parameters.names != self._cache[
            "models"
        ].parameters.names or len(self.models.parameters.free_parameters) != len(
            self._cache["models"].parameters.free_parameters
        ):
            self._to_update["models"] = self.models
            self._cache["models"] = self.models.copy()


class RayBackendMixin:
    """Ray mixin for the remote class."""

    def _get(self, name, *args, to_update={}, from_actors=False, **kwargs):
        for key, value in to_update.items():
            setattr(self, key, value)
        res = getattr(self, name)
        if isinstance(res, property):
            res = res()
        elif inspect.ismethod(res) and from_actors and "plot" not in name:
            try:
                res = res(*args, **kwargs)
            except TypeError:
                return res
        return res


class _MapDatasetActorBackend(MapDataset, RayBackendMixin):
    """MapDataset backend for parallel evaluation as a ray actor.

    Parameters
    ----------
    dataset : `MapDataset`
        MapDataset
    """

    def _from_dataset(self, dataset):
        self.__dict__.update(dataset.__dict__)
        if self.models is None:
            self.models = DatasetModels()

    def _update_stat_sum(self, values, to_update={}):
        for key, value in to_update.items():
            setattr(self, key, value)
        self.models.parameters.free_parameters.value = values
        return self.stat_sum()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import copy
import logging
from gammapy.utils.cache import CacheInstanceMixin, cachemethod
import numpy as np
from regions import PointSkyRegion
import astropy.units as u
from astropy.convolution import Tophat2DKernel
from astropy.coordinates import Angle, SkyCoord
from astropy.io import fits
from astropy.nddata import Cutout2D
from astropy.nddata.utils import overlap_slices
from astropy.utils import lazyproperty
from astropy.wcs import WCS
from astropy.wcs.utils import (
    celestial_frame_to_wcs,
    proj_plane_pixel_scales,
    wcs_to_celestial_frame,
)
from regions import RectangleSkyRegion
from gammapy.utils.array import round_up_to_even, round_up_to_odd
from gammapy.utils.compat import COPY_IF_NEEDED
from ..axes import MapAxes
from ..coord import MapCoord, skycoord_to_lonlat
from ..geom import Geom, get_shape, pix_tuple_to_idx
from ..utils import INVALID_INDEX, _check_binsz, _check_width

log = logging.getLogger(__name__)

__all__ = ["WcsGeom"]


def cast_to_shape(param, shape, dtype):
    """Cast a tuple of parameter arrays to a given shape."""
    if not isinstance(param, tuple):
        param = [param]

    param = [np.array(p, ndmin=1, dtype=dtype) for p in param]

    if len(param) == 1:
        param = [param[0].copy(), param[0].copy()]

    for i, p in enumerate(param):
        if p.size > 1 and p.shape != shape:
            raise ValueError

        if p.shape == shape:
            continue

        param[i] = p * np.ones(shape, dtype=dtype)

    return tuple(param)


def get_resampled_wcs(wcs, factor, downsampled):
    """Get resampled WCS object."""
    wcs = wcs.deepcopy()

    if not downsampled:
        factor = 1.0 / factor

    wcs.wcs.cdelt *= factor
    wcs.wcs.crpix = (wcs.wcs.crpix - 0.5) / factor + 0.5
    return wcs


class WcsGeom(Geom, CacheInstanceMixin):
    """Geometry class for WCS maps.

    This class encapsulates both the WCS transformation object and
    the image extent (number of pixels in each dimension). Provides
    methods for accessing the properties of the WCS object and
    performing transformations between pixel and world coordinates.

    Parameters
    ----------
    wcs : `~astropy.wcs.WCS`
        WCS projection object.
    npix : tuple
        Number of pixels in each spatial dimension.
    cdelt : tuple
        Pixel size in each image plane. If None then a constant pixel size will be used.
    crpix : tuple
        Reference pixel coordinate in each image plane.
    axes : list
        Axes for non-spatial dimensions.
    """

    _slice_spatial_axes = slice(0, 2)
    _slice_non_spatial_axes = slice(2, None)
    is_hpx = False
    is_region = False

    def __init__(self, wcs, npix=None, cdelt=None, crpix=None, axes=None):
        self._wcs = wcs
        self._frame = wcs_to_celestial_frame(wcs).name
        self._projection = wcs.wcs.ctype[0][5:]
        self._axes = MapAxes.from_default(axes, n_spatial_axes=2)

        if npix is None:
            if wcs.array_shape is not None:
                npix = wcs.array_shape[::-1]
            else:
                npix = (0, 0)

        if cdelt is None:
            cdelt = tuple(np.abs(self.wcs.wcs.cdelt))

        # Shape to use for WCS transformations
        wcs_shape = max([get_shape(t) for t in [npix, cdelt]])
        self._npix = cast_to_shape(npix, wcs_shape, int)
        self._cdelt = cast_to_shape(cdelt, wcs_shape, float)

        # By convention CRPIX is indexed from 1
        if crpix is None:
            crpix = tuple(1.0 + (np.array(self._npix) - 1.0) / 2.0)

        self._crpix = crpix

    @property
    def data_shape(self):
        """Shape of the `~numpy.ndarray` matching this geometry."""
        return self._shape[::-1]

    @property
    def axes_names(self):
        """All axes names."""
        return ["lon", "lat"] + self.axes.names

    @property
    def data_shape_axes(self):
        """Shape of data of the non-spatial axes and unit spatial axes."""
        return self.axes.shape[::-1] + (1, 1)

    @property
    def data_shape_image(self):
        """Shape of data of the spatial axes and unit non-spatial axes."""
        return (1,) * len(self.axes) + self.data_shape[len(self.axes) :]

    @property
    def _shape(self):
        npix_shape = tuple([np.max(self.npix[0]), np.max(self.npix[1])])
        return npix_shape + self.axes.shape

    @property
    def _shape_edges(self):
        npix_shape = tuple([np.max(self.npix[0]) + 1, np.max(self.npix[1]) + 1])
        return npix_shape + self.axes.shape

    @property
    def shape_axes(self):
        """Shape of non-spatial axes."""
        return self._shape[self._slice_non_spatial_axes]

    @property
    def wcs(self):
        """WCS projection object."""
        return self._wcs

    @property
    def frame(self):
        """Coordinate system of the projection.

        Galactic ("galactic") or Equatorial ("icrs").
        """
        return self._frame

    def cutout_slices(self, geom, mode="partial"):
        """Compute cutout slices.

        Parameters
        ----------
        geom : `WcsGeom`
            Parent geometry.
        mode : {"trim", "partial", "strict"}, optional
            Cutout slices mode. Default is "partial".

        Returns
        -------
        slices : dict
            Dictionary containing "parent-slices" and "cutout-slices".
        """
        position = geom.to_image().coord_to_pix(self.center_skydir)
        slices = overlap_slices(
            large_array_shape=geom.data_shape[-2:],
            small_array_shape=self.data_shape[-2:],
            position=[_[0] for _ in position[::-1]],
            mode=mode,
        )
        return {
            "parent-slices": slices[0],
            "cutout-slices": slices[1],
        }

    @property
    def projection(self):
        """Map projection."""
        return self._projection

    @property
    def is_allsky(self):
        """Flag for all-sky maps."""
        if np.all(np.isclose(self._npix[0] * self._cdelt[0], 360.0)) and np.all(
            np.isclose(self._npix[1] * self._cdelt[1], 180.0)
        ):
            return True
        else:
            return False

    @property
    def is_regular(self):
        """If geometry is regular in non-spatial dimensions as a boolean.

        - False for multi-resolution or irregular geometries.
        - True if all image planes have the same pixel geometry.
        """
        if self.npix[0].size > 1:
            return False
        else:
            return True

    @property
    def width(self):
        """Tuple with image dimension in degrees in longitude and latitude."""
        dlon = self._cdelt[0] * self._npix[0]
        dlat = self._cdelt[1] * self._npix[1]
        return (dlon, dlat) * u.deg

    @property
    def pixel_area(self):
        """Pixel area in deg^2."""
        # FIXME: Correctly compute solid angle for projection
        return self._cdelt[0] * self._cdelt[1]

    @property
    def npix(self):
        """Tuple with image dimension in pixels in longitude and latitude."""
        return self._npix

    @property
    def axes(self):
        """List of non-spatial axes."""
        return self._axes

    @property
    def ndim(self):
        return len(self.data_shape)

    @property
    def center_coord(self):
        """Map coordinate of the center of the geometry.

        Returns
        -------
        coord : tuple
        """
        return self.pix_to_coord(self.center_pix)

    @property
    def center_pix(self):
        """Pixel coordinate of the center of the geometry.

        Returns
        -------
        pix : tuple
        """
        return tuple((np.array(self.data_shape) - 1.0) / 2)[::-1]

    @property
    def center_skydir(self):
        """Sky coordinate of the center of the geometry.

        Returns
        -------
        pix : `~astropy.coordinates.SkyCoord`
        """
        return SkyCoord.from_pixel(self.center_pix[0], self.center_pix[1], self.wcs)

    @property
    def pixel_scales(self):
        """
        Pixel scale.

        Returns angles along each axis of the image at the CRPIX location once
        it is projected onto the plane of intermediate world coordinates.

        Returns
        -------
        angle: `~astropy.coordinates.Angle`
        """
        return Angle(proj_plane_pixel_scales(self.wcs), "deg")

    @classmethod
    def create(
        cls,
        npix=None,
        binsz=0.5,
        proj="CAR",
        frame="icrs",
        refpix=None,
        axes=None,
        skydir=None,
        width=None,
    ):
        """Create a WCS geometry object.

        Pixelization of the map is set with
        ``binsz`` and one of either ``npix`` or ``width`` arguments.
        For maps with non-spatial dimensions a different pixelization
        can be used for each image plane by passing a list or array
        argument for any of the pixelization parameters. If both npix
        and width are None then an all-sky geometry will be created.

        Parameters
        ----------
        npix : int or tuple or list, optional
            Width of the map in pixels. A tuple will be interpreted as
            parameters for longitude and latitude axes. For maps with
            non-spatial dimensions, list input can be used to define a
            different map width in each image plane. This option
            supersedes width. Default is None.
        binsz : float or tuple or list, optional
            Map pixel size in degrees. A tuple will be interpreted
            as parameters for longitude and latitude axes. For maps
            with non-spatial dimensions, list input can be used to
            define a different bin size in each image plane.
            Default is 0.5
        proj : string, optional
            Any valid WCS projection type. Default is 'CAR' (Plate-Carrée projection).
            See `WCS supported projections <https://docs.astropy.org/en/stable/wcs/supported_projections.html>`__  # noqa: E501
        frame : {"icrs", "galactic"}, optional
            Coordinate system, either Galactic ("galactic") or Equatorial ("icrs").
            Default is "icrs".
        refpix : tuple, optional
            Reference pixel of the projection. If None this will be
            set to the center of the map. Default is None.
        axes : list, optional
            List of non-spatial axes.
        skydir : tuple or `~astropy.coordinates.SkyCoord`, optional
            Sky position of map center. Can be either a SkyCoord
            object or a tuple of longitude and latitude in deg in the
            coordinate system of the map. Default is None.
        width : float or tuple or list or string, optional
            Width of the map in degrees. A tuple will be interpreted
            as parameters for longitude and latitude axes. For maps
            with non-spatial dimensions, list input can be used to
            define a different map width in each image plane.
            Default is None.


        Returns
        -------
        geom : `~WcsGeom`
            A WCS geometry object.

        Examples
        --------
        >>> from gammapy.maps import WcsGeom
        >>> from gammapy.maps import MapAxis
        >>> axis = MapAxis.from_bounds(0,1,2)
        >>> geom = WcsGeom.create(npix=(100,100), binsz=0.1)
        >>> geom = WcsGeom.create(npix=(100,100), binsz="0.1deg")
        >>> geom = WcsGeom.create(npix=[100,200], binsz=[0.1,0.05], axes=[axis])
        >>> geom = WcsGeom.create(npix=[100,200], binsz=["0.1deg","0.05deg"], axes=[axis])
        >>> geom = WcsGeom.create(width=[5.0,8.0], binsz=[0.1,0.05], axes=[axis])
        >>> geom = WcsGeom.create(npix=([100,200],[100,200]), binsz=0.1, axes=[axis])
        """
        if skydir is None:
            crval = (0.0, 0.0)
        elif isinstance(skydir, tuple):
            crval = skydir
        elif isinstance(skydir, SkyCoord):
            xref, yref, frame = skycoord_to_lonlat(skydir, frame=frame)
            crval = (xref, yref)
        else:
            raise ValueError(f"Invalid type for skydir: {type(skydir)!r}")

        if width is not None:
            width = _check_width(width)

        binsz = _check_binsz(binsz)

        shape = max([get_shape(t) for t in [npix, binsz, width]])
        binsz = cast_to_shape(binsz, shape, float)

        # If both npix and width are None then create an all-sky geometry
        if npix is None and width is None:
            width = (360.0, 180.0)

        if npix is None:
            width = cast_to_shape(width, shape, float)
            npix = (
                np.rint(width[0] / binsz[0]).astype(int),
                np.rint(width[1] / binsz[1]).astype(int),
            )
        else:
            npix = cast_to_shape(npix, shape, int)

        if refpix is None:
            nxpix = int(npix[0].flat[0])
            nypix = int(npix[1].flat[0])
            refpix = ((nxpix + 1) / 2.0, (nypix + 1) / 2.0)

        # get frame class
        frame = SkyCoord(np.nan, np.nan, frame=frame, unit="deg").frame
        wcs = celestial_frame_to_wcs(frame, projection=proj)
        wcs.wcs.crpix = refpix
        wcs.wcs.crval = crval

        cdelt = (-binsz[0].flat[0], binsz[1].flat[0])
        wcs.wcs.cdelt = cdelt

        wcs.array_shape = npix[1].flat[0], npix[0].flat[0]
        wcs.wcs.datfix()
        return cls(wcs, npix, cdelt=binsz, axes=axes)

    @property
    def footprint(self):
        """Footprint of the geometry as a `~astropy.coordinates.SkyCoord`."""
        coords = self.wcs.calc_footprint()
        return SkyCoord(coords, frame=self.frame, unit="deg")

    @property
    def footprint_rectangle_sky_region(self):
        """Footprint of the geometry as a `~regions.RectangleSkyRegion`."""
        width, height = self.width
        return RectangleSkyRegion(
            center=self.center_skydir, width=width[0], height=height[0]
        )

    @classmethod
    def from_aligned(cls, geom, skydir, width):
        """Create an aligned geometry from an existing one.

        Parameters
        ----------
        geom : `~WcsGeom`
            A reference WCS geometry object.
        skydir : tuple or `~astropy.coordinates.SkyCoord`
            Sky position of map center. Can be either a SkyCoord
            object or a tuple of longitude and latitude in degrees in the
            coordinate system of the map.
        width : float or tuple or list or string
            Width of the map in degrees. A tuple will be interpreted
            as parameters for longitude and latitude axes. For maps
            with non-spatial dimensions, list input can be used to
            define a different map width in each image plane.

        Returns
        -------
        geom : `~WcsGeom`
            An aligned WCS geometry object with specified size and center.

        """
        width = _check_width(width) * u.deg
        npix = tuple(np.round(width / geom.pixel_scales).astype(int))
        xref, yref = geom.to_image().coord_to_pix(skydir)
        xref = int(np.floor(-xref[0] + npix[0] / 2.0)) + geom.wcs.wcs.crpix[0]
        yref = int(np.floor(-yref[0] + npix[1] / 2.0)) + geom.wcs.wcs.crpix[1]
        return cls.create(
            skydir=tuple(geom.wcs.wcs.crval),
            npix=npix,
            refpix=(xref, yref),
            frame=geom.frame,
            binsz=tuple(geom.pixel_scales.deg),
            axes=geom.axes,
            proj=geom.projection,
        )

    @classmethod
    def from_header(cls, header, hdu_bands=None, format="gadf"):
        """Create a WCS geometry object from a FITS header.

        Parameters
        ----------
        header : `~astropy.io.fits.Header`
            The FITS header.
        hdu_bands : `~astropy.io.fits.BinTableHDU`, optional
            The BANDS table HDU. Default is None.
        format : {'gadf', 'fgst-ccube','fgst-template'}, optional
            FITS format convention. Default is "gadf".

        Returns
        -------
        wcs : `~WcsGeom`
            WCS geometry object.
        """
        wcs = WCS(header, naxis=2).sub(2)

        axes = MapAxes.from_table_hdu(hdu_bands, format=format)
        shape = axes.shape

        if hdu_bands is not None and "NPIX" in hdu_bands.columns.names:
            npix = hdu_bands.data.field("NPIX").reshape(shape + (2,))
            npix = (npix[..., 0], npix[..., 1])
            cdelt = hdu_bands.data.field("CDELT").reshape(shape + (2,))
            cdelt = (cdelt[..., 0], cdelt[..., 1])
        elif "WCSSHAPE" in header:
            wcs_shape = eval(header["WCSSHAPE"])
            npix = (wcs_shape[0], wcs_shape[1])
            cdelt = None
            wcs.array_shape = npix
        else:
            npix = (header["NAXIS1"], header["NAXIS2"])
            cdelt = None

        return cls(wcs, npix, cdelt=cdelt, axes=axes)

    def _make_bands_cols(self):
        cols = []
        if not self.is_regular:
            cols += [
                fits.Column(
                    "NPIX",
                    "2I",
                    dim="(2)",
                    array=np.vstack((np.ravel(self.npix[0]), np.ravel(self.npix[1]))).T,
                )
            ]
            cols += [
                fits.Column(
                    "CDELT",
                    "2E",
                    dim="(2)",
                    array=np.vstack(
                        (np.ravel(self._cdelt[0]), np.ravel(self._cdelt[1]))
                    ).T,
                )
            ]
            cols += [
                fits.Column(
                    "CRPIX",
                    "2E",
                    dim="(2)",
                    array=np.vstack(
                        (np.ravel(self._crpix[0]), np.ravel(self._crpix[1]))
                    ).T,
                )
            ]
        return cols

    def to_header(self):
        header = self.wcs.to_header()
        header.update(self.axes.to_header())
        shape = "{},{}".format(np.max(self.npix[0]), np.max(self.npix[1]))
        for ax in self.axes:
            shape += f",{ax.nbin}"

        header["WCSSHAPE"] = f"({shape})"
        return header

    @cachemethod
    def get_idx(self, idx=None, flat=False):
        pix = self.get_pix(idx=idx, mode="center")
        if flat:
            pix = tuple([p[np.isfinite(p)] for p in pix])
        return pix_tuple_to_idx(pix)

    def _get_pix_all(
        self, idx=None, mode="center", sparse=False, axis_name=("lon", "lat")
    ):
        """Get index coordinate array without footprint of the projection applied."""
        pix_all = []

        for name, nbin in zip(self.axes_names, self._shape):
            if mode == "edges" and name in axis_name:
                pix = np.arange(-0.5, nbin, dtype=float)
            else:
                pix = np.arange(nbin, dtype=float)

            pix_all.append(pix)

        # TODO: improve varying bin size coordinate handling
        if idx is not None:
            pix_all = pix_all[self._slice_spatial_axes] + [float(t) for t in idx]

        return np.meshgrid(*pix_all[::-1], indexing="ij", sparse=sparse)[::-1]

    def get_pix(self, idx=None, mode="center"):
        """Get map pixel coordinates from the geometry.

        Parameters
        ----------
        mode : {'center', 'edges'}, optional
            Get center or edge pix coordinates for the spatial axes.
            Default is "center".

        Returns
        -------
        coord : tuple
            Map pixel coordinate tuple.
        """
        pix = self._get_pix_all(idx=idx, mode=mode)
        coords = self.pix_to_coord(pix)
        m = np.isfinite(coords[0])
        for _ in pix:
            _[~m] = INVALID_INDEX.float
        return pix

    @cachemethod
    def get_coord(
        self, idx=None, mode="center", frame=None, sparse=False, axis_name=None
    ):
        """Get map coordinates from the geometry.

        Parameters
        ----------
        mode : {'center', 'edges'}, optional
            Get center or edge coordinates for the spatial axes.
            Default is "center".
        frame : str or `~astropy.coordinates.Frame`, optional
            Coordinate frame. Default is None.
        sparse : bool, optional
            Compute sparse coordinates. Default is False.
        axis_name : str, optional
            If mode = "edges", the edges will be returned for this axis.
             Default is None.

        Returns
        -------
        coord : `~MapCoord`
            Map coordinate object.
        """
        if axis_name is None:
            axis_name = ("lon", "lat")

        if frame is None:
            frame = self.frame

        pix = self._get_pix_all(idx=idx, mode=mode, sparse=sparse, axis_name=axis_name)

        data = self.pix_to_coord(pix)

        coords = MapCoord.create(
            data=data, frame=self.frame, axis_names=self.axes.names
        )
        return coords.to_frame(frame)

    def coord_to_pix(self, coords):
        coords = MapCoord.create(coords, frame=self.frame, axis_names=self.axes.names)

        if coords.size == 0:
            return tuple([np.array([]) for i in range(coords.ndim)])

        # Variable Bin Size
        if not self.is_regular:
            idxs = self.axes.coord_to_idx(coords, clip=True)
            crpix = [t[idxs] for t in self._crpix]
            cdelt = [t[idxs] for t in self._cdelt]
            pix = world2pix(self.wcs, cdelt, crpix, (coords.lon, coords.lat))
            pix = list(pix)
        else:
            pix = self._wcs.wcs_world2pix(coords.lon, coords.lat, 0)

        pix += self.axes.coord_to_pix(coords)
        return tuple(pix)

    def pix_to_coord(self, pix):
        # Variable Bin Size
        if not self.is_regular:
            idxs = pix_tuple_to_idx(pix[self._slice_non_spatial_axes])
            crpix = [t[idxs] for t in self._crpix]
            cdelt = [t[idxs] for t in self._cdelt]
            coords = pix2world(self.wcs, cdelt, crpix, pix[self._slice_spatial_axes])
        else:
            coords = self._wcs.wcs_pix2world(pix[0], pix[1], 0)

        coords = (
            u.Quantity(coords[0], unit="deg", copy=COPY_IF_NEEDED),
            u.Quantity(coords[1], unit="deg", copy=COPY_IF_NEEDED),
        )

        coords += self.axes.pix_to_coord(pix[self._slice_non_spatial_axes])
        return coords

    def pix_to_idx(self, pix, clip=False):
        pix = pix_tuple_to_idx(pix)

        idx_non_spatial = self.axes.pix_to_idx(
            pix[self._slice_non_spatial_axes], clip=clip
        )

        if not self.is_regular:
            npix = (self.npix[0][idx_non_spatial], self.npix[1][idx_non_spatial])
        else:
            npix = self.npix

        idx_spatial = []

        for idx, npix_ in zip(pix[self._slice_spatial_axes], npix):
            if clip:
                idx = np.clip(idx, 0, npix_)
            else:
                idx = np.where((idx < 0) | (idx >= npix_), -1, idx)

            idx_spatial.append(idx)

        return tuple(idx_spatial) + idx_non_spatial

    def contains(self, coords):
        idx = self.coord_to_idx(coords)
        return np.all(np.stack([t != INVALID_INDEX.int for t in idx]), axis=0)

    def to_image(self):
        return self._image_geom

    @lazyproperty
    def _image_geom(self):
        npix = (np.max(self._npix[0]), np.max(self._npix[1]))
        cdelt = (np.max(self._cdelt[0]), np.max(self._cdelt[1]))
        return self.__class__(self._wcs, npix, cdelt=cdelt)

    def to_cube(self, axes):
        npix = (np.max(self._npix[0]), np.max(self._npix[1]))
        cdelt = (np.max(self._cdelt[0]), np.max(self._cdelt[1]))
        axes = copy.deepcopy(self.axes) + axes
        return self.__class__(
            self._wcs.deepcopy(),
            npix,
            cdelt=cdelt,
            axes=axes,
        )

    def _pad_spatial(self, pad_width):
        if np.isscalar(pad_width):
            pad_width = (pad_width, pad_width)

        npix = (self.npix[0] + 2 * pad_width[0], self.npix[1] + 2 * pad_width[1])
        wcs = self._wcs.deepcopy()
        wcs.wcs.crpix += np.array(pad_width)
        cdelt = copy.deepcopy(self._cdelt)
        return self.__class__(wcs, npix, cdelt=cdelt, axes=copy.deepcopy(self.axes))

    def crop(self, crop_width):
        if np.isscalar(crop_width):
            crop_width = (crop_width, crop_width)

        npix = (self.npix[0] - 2 * crop_width[0], self.npix[1] - 2 * crop_width[1])
        wcs = self._wcs.deepcopy()
        wcs.wcs.crpix -= np.array(crop_width)
        cdelt = copy.deepcopy(self._cdelt)
        return self.__class__(wcs, npix, cdelt=cdelt, axes=copy.deepcopy(self.axes))

    def downsample(self, factor, axis_name=None):
        if axis_name is None:
            if np.any(np.mod(self.npix, factor) > 0):
                raise ValueError(
                    f"Spatial shape not divisible by factor {factor!r} in all axes."
                    f" You need to pad prior to calling downsample."
                )

            npix = (self.npix[0] / factor, self.npix[1] / factor)
            cdelt = (self._cdelt[0] * factor, self._cdelt[1] * factor)
            wcs = get_resampled_wcs(self.wcs, factor, True)
            return self._init_copy(wcs=wcs, npix=npix, cdelt=cdelt)
        else:
            if not self.is_regular:
                raise NotImplementedError(
                    "Upsampling in non-spatial axes not supported for irregular geometries"
                )
            axes = self.axes.downsample(factor=factor, axis_name=axis_name)
            return self._init_copy(axes=axes)

    def upsample(self, factor, axis_name=None):
        if axis_name is None:
            npix = (self.npix[0] * factor, self.npix[1] * factor)
            cdelt = (self._cdelt[0] / factor, self._cdelt[1] / factor)
            wcs = get_resampled_wcs(self.wcs, factor, False)
            return self._init_copy(wcs=wcs, npix=npix, cdelt=cdelt)
        else:
            if not self.is_regular:
                raise NotImplementedError(
                    "Upsampling in non-spatial axes not supported for irregular geometries"
                )
            axes = self.axes.upsample(factor=factor, axis_name=axis_name)
            return self._init_copy(axes=axes)

    def to_binsz(self, binsz):
        """Change pixel size of the geometry.

        Parameters
        ----------
        binsz : float or tuple or list
            New pixel size in degree.

        Returns
        -------
        geom : `WcsGeom`
            Geometry with new pixel size.
        """
        return self.create(
            skydir=self.center_skydir,
            binsz=binsz,
            width=self.width,
            proj=self.projection,
            frame=self.frame,
            axes=copy.deepcopy(self.axes),
        )

    def solid_angle(self):
        """Solid angle array as a `~astropy.units.Quantity` in ``sr``.

        The array has the same dimension as the WcsGeom object
        if the spatial shape is not unique along the extra axis,
        otherwise the array shape matches the spatial dimensions.

        To return solid angles for the spatial dimensions only use::

             WcsGeom.to_image().solid_angle()
        """
        return self._solid_angle

    @lazyproperty
    def _solid_angle(self):
        if self.is_regular:
            coord = self.to_image().get_coord(mode="edges").skycoord
        else:
            coord = self.get_coord(mode="edges").skycoord

        # define pixel corners
        low_left = coord[..., :-1, :-1]
        low_right = coord[..., 1:, :-1]
        up_left = coord[..., :-1, 1:]
        up_right = coord[..., 1:, 1:]

        # compute side lengths
        low = low_left.separation(low_right)
        left = low_left.separation(up_left)
        up = up_left.separation(up_right)
        right = low_right.separation(up_right)

        # compute enclosed angles
        angle_low_right = low_right.position_angle(up_right) - low_right.position_angle(
            low_left
        )
        angle_up_left = up_left.position_angle(up_right) - low_left.position_angle(
            up_left
        )

        # compute area assuming a planar triangle
        area_low_right = 0.5 * low * right * np.sin(angle_low_right)
        area_up_left = 0.5 * up * left * np.sin(angle_up_left)
        # TODO: for non-negative cdelt a negative solid angle is returned
        #  find out why and fix properly

        value = np.abs(
            u.Quantity(area_low_right + area_up_left, "sr", copy=COPY_IF_NEEDED)
        )
        if self.is_regular:
            value = value.reshape(self.data_shape_image)
        return value

    def bin_volume(self):
        """Bin volume as a `~astropy.units.Quantity`."""
        return self._bin_volume

    @lazyproperty
    def _bin_volume(self):
        """Cached property of bin volume."""
        value = self.to_image().solid_angle()

        if not self.is_image:
            value = value * self.axes.bin_volume()

        return value

    def separation(self, center):
        """Compute sky separation with respect to a given center.

        Parameters
        ----------
        center : `~astropy.coordinates.SkyCoord`
            Center position.

        Returns
        -------
        separation : `~astropy.coordinates.Angle`
            Separation angle array (2D).
        """
        coord = self.to_image().get_coord()
        return center.separation(coord.skycoord)

    def cutout(self, position, width, mode="trim", odd_npix=False, min_npix=1):
        """
        Create a cutout around a given position.

        Parameters
        ----------
        position : `~astropy.coordinates.SkyCoord`
            Center position of the cutout region.
        width : tuple of `~astropy.coordinates.Angle`
            Angular sizes of the region in (lon, lat) in that specific order.
            If only one value is passed, a square region is extracted.
        mode : {'trim', 'partial', 'strict'}, optional
            Mode option for Cutout2D, for details see `~astropy.nddata.utils.Cutout2D`.
            Default is "trim".
        odd_npix : bool, optional
            Force width to odd number of pixels.
            Default is False.
        min_npix : bool, optional
            Force width to a minimmum number of pixels.
            Default is 1.


        Returns
        -------
        cutout : `~gammapy.maps.WcsNDMap`
            Cutout map.
        """
        width = _check_width(width) * u.deg

        binsz = self.pixel_scales
        width_npix = np.clip((width / binsz).to_value(""), min_npix, None)

        if odd_npix:
            width_npix = round_up_to_odd(width_npix)

        dummy_data = np.empty(self.to_image().data_shape, dtype=bool)
        c2d = Cutout2D(
            data=dummy_data,
            wcs=self.wcs,
            position=position,
            # Cutout2D takes size with order (lat, lon)
            size=width_npix[::-1],
            mode=mode,
        )
        return self._init_copy(wcs=c2d.wcs, npix=c2d.shape[::-1])

    def boundary_mask(self, width):
        """Create a mask applying binary erosion with a given width from geometry edges.

        Parameters
        ----------
        width : tuple of `~astropy.units.Quantity`
            Angular sizes of the margin in (lon, lat) in that specific order.
            If only one value is passed, the same margin is applied in (lon, lat).

        Returns
        -------
        mask_map : `~gammapy.maps.WcsNDMap` of boolean type
            Boundary mask.

        """
        from .ndmap import WcsNDMap

        data = np.ones(self.data_shape, dtype=bool)
        return WcsNDMap.from_geom(self, data=data).binary_erode(
            width=2 * u.Quantity(width), kernel="box"
        )

    def region_mask(self, regions, inside=True):
        """Create a mask from a given list of regions.

        The mask is filled such that a pixel inside the region is filled with
        "True". To invert the mask, e.g. to create a mask with exclusion regions
        the tilde (~) operator can be used (see example below).

        Parameters
        ----------
        regions : str, `~regions.Region` or list of `~regions.Region`
            Region or list of regions (pixel or sky regions accepted).
            A region can be defined as a string ind DS9 format as well.
            See http://ds9.si.edu/doc/ref/region.html for details.
        inside : bool, optional
            For ``inside=True``, set pixels in the region to True.
            For ``inside=False``, set pixels in the region to False.
            Default is True.

        Returns
        -------
        mask_map : `~gammapy.maps.WcsNDMap` of boolean type
            Boolean region mask.


        Examples
        --------
        Make an exclusion mask for a circular region::

            from regions import CircleSkyRegion
            from astropy.coordinates import SkyCoord, Angle
            from gammapy.maps import WcsNDMap, WcsGeom

            pos = SkyCoord(0, 0, unit='deg')
            geom = WcsGeom.create(skydir=pos, npix=100, binsz=0.1)

            region = CircleSkyRegion(
                SkyCoord(3, 2, unit='deg'),
                Angle(1, 'deg'),
            )

            # the Gammapy convention for exclusion regions is to take the inverse
            mask = ~geom.region_mask([region])

        Note how we made a list with a single region,
        since this method expects a list of regions.
        """
        from gammapy.maps import Map, RegionGeom
        from gammapy.modeling.models import PointSpatialModel
        from ..region.geom import _parse_regions

        if not self.is_regular:
            raise ValueError("Multi-resolution maps not supported yet")

        regions = _parse_regions(regions)

        mask = Map.from_geom(self, data=False)
        extended_regions = []
        for reg in regions:
            if isinstance(reg, PointSkyRegion):
                spatial_model = PointSpatialModel.from_position(reg.center)
                mask.data |= spatial_model.evaluate_geom(self) > 0.0
            else:
                extended_regions.append(reg)

        if extended_regions:
            geom = RegionGeom.from_regions(extended_regions, wcs=self.wcs)
            idx = self.get_idx()
            mask.data |= geom.contains_wcs_pix(idx)

        if not inside:
            mask = ~mask

        return mask

    def region_weights(self, regions, oversampling_factor=10):
        """Compute regions weights.

        Parameters
        ----------
        regions : str, `~regions.Region` or list of `~regions.Region`
            Region or list of regions (pixel or sky regions accepted).
            A region can be defined as a string ind DS9 format as well.
            See http://ds9.si.edu/doc/ref/region.html for details.
        oversampling_factor : int, optional
            Over-sampling factor to compute the region weights.
            Default is 10.

        Returns
        -------
        map : `~gammapy.maps.WcsNDMap` of boolean type
            Weights region mask.
        """
        geom = self.upsample(factor=oversampling_factor)
        m = geom.region_mask(regions=regions)
        m.data = m.data.astype(float)
        return m.downsample(factor=oversampling_factor, preserve_counts=False)

    def binary_structure(self, width, kernel="disk"):
        """Get binary structure.

        Parameters
        ----------
        width : `~astropy.units.Quantity`, str or float
            If a float is given it interpreted as width in pixels. If an (angular)
            quantity is given it converted to pixels using ``geom.wcs.wcs.cdelt``.
            The width corresponds to radius in case of a disk kernel, and
            the side length in case of a box kernel.
        kernel : {'disk', 'box'}, optional
            Kernel shape. Default is "disk".

        Returns
        -------
        structure : `~numpy.ndarray`
            Binary structure.
        """
        width = u.Quantity(width)

        if width.unit.is_equivalent("deg"):
            width = width / self.pixel_scales

        width = round_up_to_odd(width.to_value(""))

        if kernel == "disk":
            disk = Tophat2DKernel(width[0])
            disk.normalize("peak")
            structure = disk.array
        elif kernel == "box":
            structure = np.ones(width)
        else:
            raise ValueError(f"Invalid kernel: {kernel!r}")

        shape = (1,) * len(self.axes) + structure.shape
        return structure.reshape(shape)

    def __str__(self):
        lon = self.center_skydir.data.lon.deg
        lat = self.center_skydir.data.lat.deg
        lon_ref, lat_ref = self.wcs.wcs.crval

        return (
            f"{self.__class__.__name__}\n\n"
            f"\taxes       : {self.axes_names}\n"
            f"\tshape      : {self.data_shape[::-1]}\n"
            f"\tndim       : {self.ndim}\n"
            f"\tframe      : {self.frame}\n"
            f"\tprojection : {self.projection}\n"
            f"\tcenter     : {lon:.1f} deg, {lat:.1f} deg\n"
            f"\twidth      : {self.width[0][0]:.1f} x {self.width[1][0]:.1f}\n"
            f"\twcs ref    : {lon_ref:.1f} deg, {lat_ref:.1f} deg\n"
        )

    def to_odd_npix(self, max_radius=None):
        """Create a new geometry object with an odd number of pixels and a maximum size.

        This is useful for PSF kernel creation.

        Parameters
        ----------
        max_radius : `~astropy.units.Quantity`, optional
            Maximum radius of the geometry (half the width).
            Default is None.

        Returns
        -------
        geom : `WcsGeom`
            Geometry with odd number of pixels.
        """
        if max_radius is None:
            width = self.width.max()
        else:
            width = 2 * u.Quantity(max_radius)

        binsz = self.pixel_scales.max()

        width_npix = (width / binsz).to_value("")
        npix = round_up_to_odd(width_npix)
        return WcsGeom.create(
            skydir=self.center_skydir,
            binsz=binsz,
            npix=npix,
            proj=self.projection,
            frame=self.frame,
            axes=self.axes,
        )

    def to_even_npix(self):
        """Create a new geometry object with an even number of pixels and a maximum size.

        Returns
        -------
        geom : `WcsGeom`
            Geometry with odd number of pixels.
        """
        width = self.width.max()
        binsz = self.pixel_scales.max()

        width_npix = (width / binsz).to_value("")
        npix = round_up_to_even(width_npix)
        return WcsGeom.create(
            skydir=self.center_skydir,
            binsz=binsz,
            npix=npix,
            proj=self.projection,
            frame=self.frame,
            axes=self.axes,
        )

    def is_aligned(self, other, tolerance=1e-6):
        """Check if WCS and extra axes are aligned.

        Parameters
        ----------
        other : `WcsGeom`
            Other geometry.
        tolerance : float, optional
            Tolerance for the comparison.
            Default is 1e-6.

        Returns
        -------
        aligned : bool
            Whether geometries are aligned.
        """
        for axis, otheraxis in zip(self.axes, other.axes):
            if axis != otheraxis:
                return False

        # check WCS consistency with a priori tolerance of 1e-6
        return self.wcs.wcs.compare(other.wcs.wcs, cmp=2, tolerance=tolerance)

    def is_allclose(self, other, rtol_axes=1e-6, atol_axes=1e-6, rtol_wcs=1e-6):
        """Compare two data IRFs for equivalency.

        Parameters
        ----------
        other : `WcsGeom`
            Geom to compare against.
        rtol_axes : float, optional
            Relative tolerance for the axes comparison.
            Default is 1e-6.
        atol_axes : float, optional
            Relative tolerance for the axes comparison.
            Default is 1e-6.
        rtol_wcs : float, optional
            Relative tolerance for the WCS comparison.
            Default is 1e-6.

        Returns
        -------
        is_allclose : bool
            Whether the geometry is all close.
        """
        if not isinstance(other, self.__class__):
            return TypeError(f"Cannot compare {type(self)} and {type(other)}")

        if self.data_shape != other.data_shape:
            log.debug("WcsGeom data shape is not equal")
            return False

        axes_eq = self.axes.is_allclose(other.axes, rtol=rtol_axes, atol=atol_axes)
        if not axes_eq:
            log.debug("WcsGeom axes are not equal")

        # check WCS consistency with a priori tolerance of 1e-6
        # cmp=1 parameter ensures no comparison with ancillary information
        # see https://github.com/astropy/astropy/pull/4522/files
        wcs_eq = self.wcs.wcs.compare(other.wcs.wcs, cmp=1, tolerance=rtol_wcs)
        if not wcs_eq:
            log.debug("WcsGeom wcs is not equal")
        return axes_eq and wcs_eq

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        if not (self.is_regular and other.is_regular):
            raise NotImplementedError(
                "Geom comparison is not possible for irregular geometries."
            )

        return self.is_allclose(other=other, rtol_wcs=1e-6, rtol_axes=1e-6)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self)


def pix2world(wcs, cdelt, crpix, pix):
    """Perform pixel to world coordinate transformation.

    For a WCS projection with a given pixel size (CDELT) and reference pixel
    (CRPIX). This method can be used to perform WCS transformations
    for projections with different pixelizations but the same
    reference coordinate (CRVAL), projection type, and coordinate system.

    Parameters
    ----------
    wcs : `~astropy.wcs.WCS`
        WCS transform object.
    cdelt : tuple
        Tuple of X/Y pixel size in deg. Each element should have the
        same length as ``pix``.
    crpix : tuple
        Tuple of reference pixel parameters in X and Y dimensions. Each
        element should have the same length as ``pix``.
    pix : tuple
        Tuple of pixel coordinates.
    """
    pix_ratio = [
        np.abs(wcs.wcs.cdelt[0] / cdelt[0]),
        np.abs(wcs.wcs.cdelt[1] / cdelt[1]),
    ]
    pix = (
        (pix[0] - (crpix[0] - 1.0)) / pix_ratio[0] + wcs.wcs.crpix[0] - 1.0,
        (pix[1] - (crpix[1] - 1.0)) / pix_ratio[1] + wcs.wcs.crpix[1] - 1.0,
    )
    return wcs.wcs_pix2world(pix[0], pix[1], 0)


def world2pix(wcs, cdelt, crpix, coord):
    pix_ratio = [
        np.abs(wcs.wcs.cdelt[0] / cdelt[0]),
        np.abs(wcs.wcs.cdelt[1] / cdelt[1]),
    ]
    pix = wcs.wcs_world2pix(coord[0], coord[1], 0)
    return (
        (pix[0] - (wcs.wcs.crpix[0] - 1.0)) * pix_ratio[0] + crpix[0] - 1.0,
        (pix[1] - (wcs.wcs.crpix[1] - 1.0)) * pix_ratio[1] + crpix[1] - 1.0,
    )
